
__all__ = ["process_cbar", "make_cols_from_cmap", "process_axes",
           "kde", "make_clrs_from_cmap", "map_array_to_cmap", "AddMarginalPlots"]

from typing import Union, Any, Optional, Tuple, List
from collections.abc import KeysView, ValuesView

import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.path import Path
from matplotlib.spines import Spine
from matplotlib.colors import Normalize
from matplotlib.transforms import Affine2D
from matplotlib.projections.polar import PolarAxes
from matplotlib.patches import RegularPolygon
from matplotlib.projections import register_projection
from mpl_toolkits.axes_grid1 import make_axes_locatable


BAR_CMAPS = ['Blues', 'BuGn', 'gist_earth_r',
             'GnBu', 'PuBu', 'PuBuGn', 'summer_r']


def process_axes(
        ax: plt.Axes=None,
        label:str = None,
        legend_kws:dict = None,
        logy:bool = False,
        logx: bool= False,
        xlabel: str = None,
        xlabel_kws:dict=None,
        xtick_kws:dict = None,
        ylim:tuple = None,
        ylabel:str = None,
        ylabel_kws:dict = None,
        ytick_kws:dict = None,
        show_xaxis: bool = True,
        show_yaxis:bool = True,
        top_spine=None,
        bottom_spine=None,
        right_spine=None,
        left_spine=None,
        invert_yaxis: bool = False,
        max_xticks=None,
        min_xticks=None,
        title:str = None,
        title_kws:dict=None,
        grid=None,
        grid_kws:dict = None,
        tight_layout:bool = False,
)-> plt.Axes:
    """
    processing of matplotlib Axes

    Parameters
    ----------
    ax : plt.Axes
        the :obj:`matplotlib.axes` axes which needs to be processed.
    label : str (default=None)
        will be used for legend
    legend_kws : dict, optional
        dictionary of keyword arguments to :obj:`matplotlib.axes.Axes.legend`
        These include loc, fontsize, bbox_to_anchor, markerscale
    logy : bool
        whether to convert y-axes to logrithmic scale or not
    logx : bool
        whether to convert x-axes to logrithmic scale or not
    xlabel : str
        label for x-axes
    xlabel_kws : dict
        keyword arguments for :obj:`matplotlib.axes.Axes.set_xlabel`
        as ax.set_xlabel(xlabel, **xlabel_kws)
    xtick_kws :
        # for axes.tick_params such as which, labelsize, colors etc
    min_xticks : int
        maximum number of ticks on x-axes
    max_xticks : int
        minimum number of ticks on x-axes
    ylabel : str
        label for y-axes
    ylabel_kws : dict
        ylabel kwargs for :obj:`matplotlib.axes.Axes.set_ylabel`
    ytick_kws :
        for axes.tick_params()  such as which, labelsize, colors etc
    ylim :
        limit for y axes
    invert_yaxis :
        whether to invert y-axes or not. It true following command will be
        executed ax.set_ylim(ax.get_ylim()[::-1])
    title : str
        title for axes :obj:`matplotlib.axes.Axes.set_title`
    title_kws : dict
        title kwargs
    grid :
        will be fed to ax.grid(grid,...)
    grid_kws : dict
        dictionary of keyword arguments for ax.grid(grid, **grid_kws)
    left_spine :
    right_spine :
    top_spine :
    bottom_spine :
    show_xaxis : bool, optional (default=True)
        whether to show x-axes or not
    show_yaxis : bool, optional (default=True)
        whether to show y-axes or not
    tight_layout : bool (default=False)
        whether to execulte plt.tight_layout() or not

    Returns
    -------
    plt.Axes
        the matplotlib Axes object :obj:`matplotlib.axes` which was passed to
        this function
    """
    if ax is None:
        ax = plt.gca()

    if label:
        if label != "__nolabel__":
            legend_kws = legend_kws or {}
            ax.legend(**legend_kws)

    if ylabel:
        ylabel_kws = ylabel_kws or {}
        ax.set_ylabel(ylabel, **ylabel_kws)

    if logy:
        ax.set_yscale('log')
    if logx:
        ax.set_xscale('log')

    if invert_yaxis:
        ax.set_ylim(ax.get_ylim()[::-1])

    if ylim:
        ax.set_ylim(ylim)

    if xlabel:  # better not change these paras if user has not defined any x_label
        xtick_kws = xtick_kws or {}
        ax.tick_params(axis="x", **xtick_kws)

    if ylabel:
        ytick_kws = ytick_kws or {}
        ax.tick_params(axis="y", **ytick_kws)

    ax.get_xaxis().set_visible(show_xaxis)

    if xlabel:
        xlabel_kws = xlabel_kws or {}
        ax.set_xlabel(xlabel, **xlabel_kws)

    if top_spine is not None:
        ax.spines['top'].set_visible(top_spine)
    if bottom_spine is not None:
        ax.spines['bottom'].set_visible(bottom_spine)
    if right_spine is not None:
        ax.spines['right'].set_visible(right_spine)
    if left_spine is not None:
        ax.spines['left'].set_visible(left_spine)

    if max_xticks is not None:
        min_xticks = min_xticks or max_xticks-1
        assert isinstance(min_xticks, int)
        assert isinstance(max_xticks, int)
        loc = mdates.AutoDateLocator(minticks=min_xticks, maxticks=max_xticks)
        ax.xaxis.set_major_locator(loc)
        fmt = mdates.AutoDateFormatter(loc)
        ax.xaxis.set_major_formatter(fmt)

    if title:
        title_kws = title_kws or {}
        ax.set_title(title, **title_kws)

    if grid:
        grid_kws = grid_kws or {}
        ax.grid(grid, **grid_kws)

    if not show_yaxis:
        ax.get_yaxis().set_visible(False)

    if tight_layout:
        plt.tight_layout()

    return ax


def make_cols_from_cmap(
        cm: str,
        num_cols: int,
        low=0.0,
        high=1.0
)->np.ndarray:
    """make rgb colors from a color pallete"""
    cols = getattr(plt.cm, cm)(np.linspace(low, high, num_cols))
    return cols


def to_1d_array(array_like) -> np.ndarray:
    """returned array has shape (n,) """
    if array_like.__class__.__name__ in ['list', 'tuple', 'Series']:
        return np.array(array_like)

    elif array_like.__class__.__name__ == 'ndarray':
        if array_like.ndim == 1:
            return array_like
        else:
            assert array_like.size == len(array_like), f"""
            cannot convert multidim array of shape {array_like.shape} to 1d"""
            return array_like.reshape(-1, )

    elif array_like.__class__.__name__ == 'DataFrame' and array_like.ndim == 2:
        assert len(array_like) == array_like.size
        return array_like.values.reshape(-1,)

    elif isinstance(array_like, (float, int)):
        return np.array([array_like])

    elif isinstance(array_like, (KeysView, ValuesView)):
        return np.array(list(array_like))
    else:
        try:
            array = np.array(array_like)
            assert len(array) == array.size
        except Exception:
            raise ValueError(f"""
            cannot convert object {array_like.__class__.__name__}  to 1d """)
        return array

def has_multi_cols(data)->bool:
    """returns True if data contains multiple columns"""
    if hasattr(data, "ndim") and hasattr(data, "shape"):
        if data.ndim == 2 and data.shape[1]>1:
            return True
    return False


def kde(
        y:np.ndarray,
        bw_method:str = "scott",
        bins:int = 1000,
        cut:Union[float, Tuple[float]] = 0.5,
)->Tuple[Union[np.ndarray, Tuple[np.ndarray, Optional[float]]], Any]:
    """
    Generate Kernel Density Estimate plot using Gaussian kernels.

    parameters
    ----------
    y :
        array like
    bw_method :
        method to calculate
    bins :
        number of bins
    cut :
    """

    # don't want to make whole easy_mpl dependent upon scipy
    from scipy.stats import gaussian_kde

    if isinstance(cut, float):
        cut = (cut, cut)

    y = np.array(y)
    if 'int' not in y.dtype.name:  # 'object' types have problem in removing nans
        y = np.array(y, dtype=np.float32)

    assert len(y) == y.size
    y = y[~np.isnan(y)]
    gkde = gaussian_kde(y.reshape(-1,), bw_method=bw_method)

    sample_range = np.nanmax(y) - np.nanmin(y)
    ind = np.linspace(
        np.nanmin(y) - cut[0] * sample_range,
        np.nanmax(y) + cut[1] * sample_range,
        bins,
    )

    return ind, gkde.evaluate(ind)


def annotate_imshow(
        im,
        data:np.ndarray=None,
        textcolors:Union[tuple, np.ndarray]=("black", "white"),
        threshold=None,
        fmt = '%.2f',
        **text_kws
):
    """annotates imshow
https://matplotlib.org/stable/gallery/images_contours_and_fields/image_annotated_heatmap.html
    """

    if data is None:
        data = im.get_array()

    use_threshold = True
    if isinstance(textcolors, np.ndarray) and textcolors.shape == data.shape:
        assert threshold is None, f"""
        if textcolors is given as array then threshold should be None"""
        use_threshold = False
    else:
        if threshold is not None:
            threshold = im.norm(threshold)
        else:
            threshold = im.norm(data.max()) / 2

    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            s = fmt % float(data[i, j])
            if use_threshold:
                _ = im.axes.text(j, i, s,
                        color=textcolors[int(im.norm(data[i, j]) > threshold)],
                        **text_kws)
            else:
                _ = im.axes.text(j, i, s,
                        color=textcolors[i, j],
                        **text_kws)
    return


def register_projections(num_vars, frame="polygon", grids="polygon"):
    """
    Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {'circle', 'polygon'}
        Shape of frame surrounding axes.
    grids: {"circle", "polygon"}

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2*np.pi, num_vars, endpoint=False)

    class RadarTransform(PolarAxes.PolarTransform):

        def transform_path_non_affine(self, path):
            # Paths with non-unit interpolation steps correspond to gridlines,
            # in which case we force interpolation (to defeat PolarTransform's
            # autoconversion to circular arcs).
            if path._interpolation_steps > 1:
                path = path.interpolated(num_vars)
            return Path(self.transform(path.vertices), path.codes)

    class RadarAxes(PolarAxes):

        name = 'radar'
        # use 1 line segment to connect specified points
        RESOLUTION = 1
        PolarTransform = RadarTransform

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # rotate plot such that the first axis is at the top
            self.set_theta_zero_location('N')

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default"""
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            # FIXME: markers at x[0], y[0] get doubled-up
            if x[0] != x[-1]:
                x = np.append(x, x[0])
                y = np.append(y, y[0])
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            # The Axes patch must be centered at (0.5, 0.5) and of radius 0.5
            # in axes coordinates.
            if frame == 'circle':
                pass #return Circle((0.5, 0.5), 0.5)
            elif frame == 'polygon':
                return RegularPolygon((0.5, 0.5), num_vars,
                                      radius=.5, edgecolor="k")
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

        def _gen_axes_spines(self):
            if frame == 'circle':
                return super()._gen_axes_spines()
            elif frame == 'polygon':
                # spine_type must be 'left'/'right'/'top'/'bottom'/'circle'.
                spine = Spine(axes=self,
                              spine_type='circle',
                              path=Path.unit_regular_polygon(num_vars))
                # unit_regular_polygon gives a polygon of radius 1 centered at
                # (0, 0) but we want a polygon of radius 0.5 centered at (0.5,
                # 0.5) in axes coordinates.
                spine.set_transform(Affine2D().scale(.5).translate(.5, .5)
                                    + self.transAxes)
                return {'polar': spine}
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

    register_projection(RadarAxes)
    return theta


def _rescale(y:np.ndarray, _min=0.0, _max=1.0)->np.ndarray:
    """rescales the y array between _min and _max"""
    y_std = (y - np.min(y, axis=0)) / (np.max(y, axis=0) - np.min(y, axis=0))

    return y_std * (_max - _min) + _min


def version_info():
    import matplotlib
    from . import __version__
    info = dict()
    info['easy_mpl'] = __version__
    info['matplotlib'] = matplotlib.__version__
    info['numpy'] = np.__version__

    try:
        import pandas as pd
        info['pandas'] = pd.__version__
    except Exception:
        pass

    return info


def is_dataframe(obj)->bool:
    if all([hasattr(obj, attr) for attr in ["columns", "index", "values", "shape"]]):
        return True
    return False


def is_series(obj)->bool:
    if all([hasattr(obj, attr) for attr in ["name", "index", "values"]]):
        return True
    return False


def create_subplots(
        naxes:int,
        ax:plt.Axes = None,
        figsize:tuple = None,
        ncols:int=None,
        **fig_kws
)->Tuple:
    """
    creates the subplots. If naxes is 1 and ax is None, then current available axes
    is returned using plt.gca(). If naxes is 1 and ax is not None, then ax is returned
    as it is. If naxes > 1 and ax is None, then new axes are created. If naxes > 1 and
    ax is not None, then ax is closed with a warning and new axes' are created.
    If naxes are > 1, then nrows and ncols are calculated automatically unless cols
    is given. The last redundent axes are switched off in case naxes < (nrows*ncols).
    """
    if ax is None:

        if naxes == 1:
            # if we need just one axes, just get the currently available one
            ax = plt.gca()
            fig = ax.get_figure()

        else:
            nrows, ncols = get_layout(naxes, ncols=ncols)
            plt.close('all')
            fig, ax = plt.subplots(nrows, ncols, figsize=figsize, **fig_kws)
            switch_off_redundant_axes(naxes, nrows * ncols, ax)

    else:
        fig = ax.get_figure()
        if naxes == 1:
            return fig, ax

        nrows, ncols = get_layout(naxes, ncols=ncols)
        plt.close('all')
        fig, ax = plt.subplots(nrows, ncols, figsize=figsize, **fig_kws)
        switch_off_redundant_axes(naxes, nrows*ncols, ax)

    return fig, ax


def switch_off_redundant_axes(naxes, nplots, axarr):

    shape = axarr.shape
    axarr = axarr.flatten()
    if naxes != nplots:
        for ax in axarr[naxes:]:
            ax.set_visible(False)

    return axarr.reshape(shape)


def get_layout(naxes:int, ncols:int=None)->tuple:

    if ncols:
        if ncols==1:
            nrows = naxes
            return nrows, ncols
        else:
            nrows = naxes-ncols
            return nrows, ncols

    layouts = {1: (1, 1), 2: (1, 2), 3: (2, 2), 4: (2, 2)}
    try:
        nrows, ncols = layouts[naxes]
    except KeyError:
        k = 1
        while k ** 2 < naxes:
            k += 1

        if (k - 1) * k >= naxes:
            nrows, ncols = k, (k - 1)
        else:
            nrows, ncols =  k, k
    return nrows, ncols


class AddMarginalPlots(object):
    """
    Adds marginal plots for an axes.

    parameters
    -----------
    ax : plt.Axes
        :obj:`matplotlib.axes` on which to add the marginal plots
    pad : float
        float or tuple of two. Distance between main axes and marginal axes
    size :
        width and height of marginal axes
    hist : bool (default=True)
        whether to draw histogram on marginal axes or not
    hist_kws : dict
        keyword arguments that will go to axes.hist function for drawing
        histograms on maginal plots. It can be a single dictionary or
        a list of two dictionaries. By default following values are used
        "linewidth":0.5, "edgecolor":"k"
    ridge : bool (default=True)
        whether to draw ridge line or not
    ridge_line_kws : dict
        keyword arguments that will go to axes.plot function for drawing
        ridge line on maginal plots. It can
        be a single dictionary or a list of two dictionaries
    fill_kws : dict
        keyword arguments that will go to axes.fill_between function fill
        between ridge line on maginal plots. Only valid if ``hist`` is False.
        It can be a single dictionary or a list of two dictionaries
    fix_limits : bool
        If true, then will not shrink/expand the axes after drawing hist/ridge line

    Examples
    ---------
    >>> from easy_mpl import plot
    >>> x = np.random.normal(size=100)
    >>> y = np.random.normal(size=100)
    >>> e = x-y
    >>> ax = plot(e, show=False)
    >>> AddMarginalPlots(ax, hist=True)(x, y)
    >>> plt.show()

    See :ref:`sphx_glr_auto_examples_marginal_plots.py` for more examples
    """
    def __init__(
            self,
            ax:plt.Axes,
            pad:float=0.25,
            size:float = 0.7,
            hist:bool = True,
            hist_kws:Union[dict, List[dict]] = None,
            ridge:bool = True,
            ridge_line_kws:Union[dict, List[dict]] = None,
            fill_kws:Union[dict, List[dict]] = None,
            fix_limits:bool = True
    ):

        self.ax = ax

        if not isinstance(pad, (list, tuple)):
            pad = [pad, pad]
        self.pad = pad

        if not isinstance(size, (list, tuple)):
            size = [size, size]
        self.size = size

        self.hist = hist
        self.ridge = ridge

        self.HIST_KWS = self.verify_kws(hist_kws)
        self.ridge_line_kws = self.verify_kws(ridge_line_kws)
        self.fill_kws = self.verify_kws(fill_kws)

        self.fix_limits = fix_limits

    def __call__(
            self,
            x,
            y,
            top_axes:plt.Axes=None,
            right_axes:plt.Axes=None
    )->tuple:
        """
        draws the top and right marginal axes.

        Parameters
        -----------
        x : array like
            the array whose distribution is to be shown on top of
            main axes
        y : array like
            the array whose distribution is to be shown on right side of
            main axes
        top_axes : plt.Axes
            the axes on which distribution of y is shown as histogram/ridge.
            If not given, will be created
        right_axes : plt.Axes
            the axes on which distribution of y is shown as histogram/ridge
            If not given, will be created using matplotlib.make_axes_locatable().append()
        """
        if top_axes is not None:
            assert isinstance(top_axes, plt.Axes)
            assert isinstance(right_axes, plt.Axes)
        else:
            self.divider = make_axes_locatable(self.ax)

        axHistx = self.add_ax_marg_x(x, hist_kws=self.HIST_KWS[0], ax=top_axes)
        axHisty = self.add_ax_marg_y(y_data=y, hist_kws=self.HIST_KWS[1],
                                     ax=right_axes)

        # make some labels invisible
        plt.setp(axHistx.get_xticklabels() + axHisty.get_yticklabels(),
                 visible=False)

        despine_axes(self.ax, keep=["left", "bottom"])

        return axHistx, axHisty

    def add_ax_marg_x(
            self,
            x_data,
            hist_kws:dict=None,
            ax:plt.Axes = None,
    ):
        """Adds the axes on top of main axes"""
        line_kws = self._get_line_kws(self.ridge_line_kws[0])

        if self.fix_limits:
            xlim = np.array(self.ax.get_xlim()) * 1.05

        _hist_kws = {"linewidth":0.5, "edgecolor":"k"}
        if hist_kws is not None:
            _hist_kws.update(hist_kws)

        if ax is None:
            new_axes = self.divider.append_axes("top", self.size[0],
                                       pad=self.pad[0], sharex=self.ax)
        else:
            new_axes = ax

        despine_axes(new_axes, keep="bottom")
        new_axes.set_yticks([])

        if self.hist:

            new_axes.hist(x_data, **_hist_kws)
            if self.ridge:
                # we draw histogram on old axes so that kde line
                # comes on top
                ax2 = new_axes.twinx()
                despine_axes(ax2)
                ax2.set_yticks([])
        else:
            ax2 = new_axes

        if self.ridge:
            ind, data = kde(x_data, cut=0.2)
            ax2.plot(ind, data, **line_kws)

        if not self.hist and self.ridge:
            # filling is only done when ridge is drawn and historgram is not
            fill_kws = self._get_fill_kws(self.fill_kws[0], n=len(ind))

            ax2.fill_between(ind, data, **fill_kws)

        if self.fix_limits:
            self.ax.set_xlim(*xlim.tolist())

        return new_axes

    def add_ax_marg_y(
            self,
            y_data,
            hist_kws: dict = None,
            ax:plt.Axes = None
    ):
        """Adds the axes on right side of main axes"""
        line_kws = self._get_line_kws(self.ridge_line_kws[1])

        if self.fix_limits:
            ylim = self.ax.get_ylim()

        _hist_kws = {"linewidth":0.5,
                     "edgecolor":"k",
                     'orientation':'horizontal'}
        if hist_kws is not None:
            _hist_kws.update(hist_kws)

        if ax is None:
            new_axes = self.divider.append_axes(
                "right",
                self.size[1],
                pad=self.pad[1],
                sharey=self.ax)
        else:
            new_axes = ax

        despine_axes(new_axes, keep="left")
        new_axes.set_xticks([])

        if self.hist:

            new_axes.hist(y_data, **_hist_kws)

            if self.ridge:
                # create a twin axes for ridge so that it comes on top
                ax2 = new_axes.twiny()
                despine_axes(ax2)
                ax2.set_xticks([])
        else:
            ax2 = new_axes

        if self.ridge:
            ind, data = kde(y_data, cut=0.2)
            ax2.plot(data, ind, **line_kws)

        if not self.hist and self.ridge:
            # filling is only done when ridge is drawn and historgram is not
            fill_kws = self._get_fill_kws(self.fill_kws[1], n=len(ind))
            ax2.fill_betweenx(ind, data, **fill_kws)

        if self.fix_limits:
            self.ax.set_ylim(ylim)

        return new_axes

    @staticmethod
    def _get_line_kws(line_kws:dict)->dict:
        _line_kws = {'color': 'k', 'lw': 1.0}
        if line_kws is not None:
            _line_kws.update(line_kws)
        return _line_kws

    @staticmethod
    def _get_fill_kws(fill_kws:dict, n:int)->dict:
        _fill_kws = {"alpha": 0.5,
                     'where': np.array([True for _ in range(n)])
                     }
        if fill_kws is not None:
            _fill_kws.update(fill_kws)
        return _fill_kws

    @staticmethod
    def verify_kws(kws:Union[dict, List[dict]]=None)->list:
        if kws is not None and not isinstance(kws, list):
            assert isinstance(kws, dict)
            kws = [kws, kws]

        elif kws is None:
            kws = [None, None]

        assert len(kws) == 2
        return kws


def despine_axes(axes, keep=None):

    if not isinstance(keep, list):
        keep = [keep]

    spines = ["top", "bottom", "right", "left"]
    for s in spines:
        if s not in keep:
            axes.spines[s].set_visible(False)
    return


def make_clrs_from_cmap(*args, **kwargs):
    return make_cols_from_cmap(*args, **kwargs)


def is_rgb(color)->bool:
    """returns True of ``color`` is rgb else returns False"""
    cond_1 = isinstance(color, (list, np.ndarray, tuple))
    if cond_1  and len(color) in [3,4] and isinstance(color[0], (int, float)):
        return True
    return False


def map_array_to_cmap(array, cmap:str, clip:bool = True)->tuple:
    """creates cmap using values in array

    Parameters
    ------------
    array :
        array like
    cmap : str
    clip : bool
    """
    norm = Normalize(vmin=np.nanmin(array).item(),
                     vmax=np.nanmax(array).item(), clip=clip)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmap)
    colors = mapper.to_rgba(array)
    return colors, mapper


def process_cbar(
        ax,
        mappable,
        border:bool = True,
        width: Union[int, float] = None,
        pad: Union[int, float] = 0.2,
        orientation:str = "vertical",
        title:str = None,
        title_kws: dict = None
):
    """
    makes and processes colobar to an axisting axes

    Parameters
    ----------
    ax : plt.Axes
    mappable :
    border : bool
        wether to draw the border or not
    pad :
    orientation : str
    title : str
    title_kws : dict
        ``rotation``
        ``labelpad``
        ``fontsize``
        ``weight``

    Returns
    -------
    plt.colorbar
    """
    if orientation == "vertical":
        position = "right"
    else:
        position = "bottom"

    cb_params = {'pad': pad, 'orientation': orientation}
    divider = make_axes_locatable(ax)
    cax = divider.append_axes(position, size="5%", pad=pad)
    cbar = plt.colorbar(mappable, cax=cax, **cb_params)

    if not border:
        # Turn spines off and create white grid.
        if isinstance(cax.spines, dict):
            for sp in cax.spines:
                cax.spines[sp].set_visible(False)
        else:
            cax.spines[:].set_visible(False)

    if title:
        if title_kws is None: title_kws = {}

        if orientation == "vertical":
            cbar.ax.set_ylabel(title, title_kws)
        else:
            cbar.ax.set_xlabel(title, **title_kws)
    return cbar




__all__ = ["dumbbell_plot"]

from typing import Tuple

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.lines as mlines

from .utils import is_rgb
from .utils import make_clrs_from_cmap
from .utils import to_1d_array, process_axes
from ._scatter import scatter


def dumbbell_plot(
        start,
        end,
        labels=None,
        line_color = None,
        start_marker_color = None,
        end_marker_color = None,
        start_kws: dict = None,
        end_kws: dict = None,
        line_kws: dict = None,
        sort_start:str = None,
        sort_end:str = None,
        ax: plt.Axes = None,
        ax_kws:dict = None,
        show: bool = True
) -> Tuple[plt.Axes, mpl.collections.PathCollection, mpl.collections.PathCollection]:
    """
    Dumbell plot which indicates variation of several variables
    from start to end.

    Parameters
    ----------
        start : list, array, series
            an array consisting of starting values
        end : list, array, series
            an array consisting of end values
        labels : list, array, series, optional
            names of values in start/end arrays. It is used to label
            ticklabcls on y-axis
        line_color :
            color for lines.  This can be a color name, rbg value, array of rbg values
            for each marker or a color palette name. This can be used to have separate
            color for a each line.
        start_marker_color :
            color for starting markers. This can be a color name, rbg value, array
            of rbg values for each marker or a color palette name. This can be
            used to have separate color for a each marker.
        end_marker_color :
            color for end markers. T This can be a color name, rbg value, array of
            rbg values for each marker or a color palette name. his can be used to
            have separate color for a each marker.
        start_kws : dict, optional
            any additional keyword arguments for :py:func:`easy_mpl.utils.scatter` to modify start
            markers such as ``color``, ``label`` etc
        end_kws : dict, optional
            any additional keyword arguments for :py:func:`easy_mpl.utils.scatter` to modify end
            markers such as ``color``, ``label`` etc
        line_kws : dict, optional
            any additional keyword arguments for `lines.Line2D`_ to modify line
            style/color which connects dumbbells.
        sort_start : str (default=None)
            either "ascend" or "descend"
        sort_end : str (default=None)
            either "ascend" or "descend"
        ax : plt.Axes, optional
            matplotlib axes object to work with. If not given then currently available
            axes will be used.
        ax_kws : dict optional
            any keyword arguments for :py:func:`easy_mpl.utils.process_axes`.
        show : bool, optional
            whether to show the plot or not

    Returns
    -------
    axes :
        :obj:`matplotlib.axes` matplotlib axes object on which dumbells are drawn
    st_pc :
        :obj:`matplotlib.collections.PathCollection`
    en_pc :
        :obj:`matplotlib.collections.PathCollection`

    Examples
    --------
        >>> import numpy as np
        >>> from easy_mpl import dumbbell_plot
        >>> st = np.random.randint(1, 5, 10)
        >>> en = np.random.randint(11, 20, 10)
        >>> dumbbell_plot(st, en)
        ... # modify line color
        >>> dumbbell_plot(st, en, line_kws={'color':"black"})

    See :ref:`sphx_glr_auto_examples_dumbell.py` for more examples

    .. _lines.Line2D:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.lines.Line2D.html

    """

    if ax_kws is None:
        ax_kws = dict()

    if ax is None:
        ax = plt.gca()
        if 'figsize' in ax_kws:
            figsize = ax_kws.pop('figsize')
            ax.figure.set_size_inches(figsize)

    # convert starting and ending values to 1d array
    start = to_1d_array(start)
    end = to_1d_array(end)

    index = np.arange(len(start))

    assert len(start) == len(end) == len(index)

    if labels is None:
        labels = np.arange(len(index))

    _line_kws = {'color': 'skyblue'}
    if line_kws is not None:
        _line_kws.update(line_kws)

    line_colors = _get_color(line_color, _line_kws, len(start))

    # assigning colors
    _start_kws = {'color': '#a3c4dc', "label": "Start"}
    if start_kws:
        _start_kws.update(start_kws)

    _end_kws = {'color': '#0e668b', "label": "End"}
    if end_kws:
        _end_kws.update(end_kws)

    st_mc_colors = _get_color(start_marker_color, _start_kws, len(start))
    en_mc_colors = _get_color(end_marker_color, _end_kws, len(start))

    if sort_start:
        start, end, labels, line_colors, st_mc_colors, en_mc_colors = _handle_sort(
            sort_start, start, start, end, labels, line_colors, st_mc_colors, en_mc_colors)
    elif sort_end:
        start, end, labels, line_colors, st_mc_colors, en_mc_colors  = _handle_sort(
            sort_end, end, start, end, labels, line_colors, st_mc_colors, en_mc_colors)

    # draw line segment
    def line_segment(p1, p2, axes, color):
        l = mlines.Line2D([p1[0], p2[0]], [p1[1], p2[1]], color=color, **_line_kws)
        axes.add_line(l)
        return

    # joining points together using line segments
    for (_idx, idx), _p1, _p2 in zip(enumerate(index), end, start):
        line_segment([_p1, idx], [_p2, idx], ax, color=line_colors[_idx])

    # circles are plotted after line so that lines don't enter inside the circles

    # plotting points for starting and ending values
    ax, st_paths = scatter(y=index, x=start, show=False, ax=ax,
                    color=st_mc_colors, **_start_kws)
    ax, en_paths = scatter(y=index, x=end, ax=ax, show=False,
                    color=en_mc_colors, **_end_kws)

    ax.legend()

    # set labels
    ax.set_yticks(index)
    ax.set_yticklabels(labels)

    if ax_kws:
        process_axes(ax=ax, **ax_kws)

    # show plot if show=True
    if show:
        plt.show()

    return ax, st_paths, en_paths


def _get_color(sugg_clr, kws, n)->list:

    if sugg_clr is None:
        colors = [kws['color'] for _ in range(n)]
    elif isinstance(sugg_clr, str):
        if sugg_clr in plt.colormaps():
            # todo
            #  this will result in wrong colorbar if these colors are used
            #  in for plot/scatter
            colors = make_clrs_from_cmap(sugg_clr, n, 0.1, 0.9)
        else:  # 'k'
            colors = [sugg_clr for _ in range(n)]
    elif is_rgb(sugg_clr):
        colors = [sugg_clr for _ in range(n)]
    else:
        assert hasattr(sugg_clr, '__len__') and len(sugg_clr) == n, f"Invalid color {sugg_clr}"
        colors = sugg_clr

    kws.pop('color')

    return colors


def _handle_sort(sort_type, sort_wrt, start, end, labels,
                 line_colors, st_mc_clr, en_mc_clr):

    assert sort_type in ["ascend", "descend"]
    if sort_type == "ascend":
        sort_idx = np.argsort(sort_wrt)
    else:
        sort_idx = np.flip(np.argsort(sort_wrt))

    start = np.array(start)[sort_idx]
    end = np.array(end)[sort_idx]
    labels = np.array(labels)[sort_idx]

    if not isinstance(line_colors, str):
        line_colors = np.array(line_colors)[sort_idx]

    if not isinstance(st_mc_clr, str):
        st_mc_clr = np.array(st_mc_clr)[sort_idx]

    if not isinstance(en_mc_clr, str):
        en_mc_clr = np.array(en_mc_clr)[sort_idx]

    return start, end, labels, line_colors, st_mc_clr, en_mc_clr



__all__ = ["parallel_coordinates"]


from typing import Union

import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches

from .utils import _rescale


def parallel_coordinates(
        data: Union[np.ndarray,],
        categories: Union[np.ndarray, list] = None,
        names: list = None,
        cmap: str = None,
        linestyle: str = "bezier",
        coord_title_kws: dict = None,
        title: str = 'Parallel Coordinates Plot',
        figsize: tuple = None,
        ticklabel_kws: dict = None,
        show: bool = True
) -> plt.Axes:
    """
    parallel coordinates plot
    modifying after https://stackoverflow.com/a/60401570/5982232

    Parameters
    ----------
        data : array, DataFrame
            a two dimensional array with the shape (rows, columns). It can also
            be pandas DataFrame
        categories : list, array
            1 dimensional array which contain class labels of the of each row in
            data. It can be either categorical or continuous numerical values.
            If not given, colorbar will not be drawn. The length of categroes
            array must be equal to length of/rows in data.
        names : list, optional
            Labels for columns in data. It's length should be equal to number of
            oclumns in data.
        cmap : str, optional
            colormap to be used
        coord_title_kws : dict, optional
            keyword arguments for coodinate titles. All of these arguments will go to
            :obj:`matplotlib.axes.Axes.set_xticklabels`
        linestyle : str, optional
            either "straight" or "bezier". Default is "bezier".
        title : str, optional
            title for the Figure
        figsize : tuple, optional
            figure size
        ticklabel_kws : dict, optional
            keyword arguments for ticklabels on y-axis
        show : bool, optional
            whether to show the plot or not

    Returns
    -------
        matplotlib Axes

    Examples
    --------
    >>> import random
    >>> import numpy as np
    >>> import pandas as pd
    >>> from easy_mpl import parallel_coordinates
    ...
    >>> ynames = ['P1', 'P2', 'P3', 'P4', 'P5']  # feature/column names
    >>> N1, N2, N3 = 10, 5, 8
    >>> N = N1 + N2 + N3
    >>> categories_ = ['a', 'b', 'c', 'd', 'e', 'f']
    >>> y1 = np.random.uniform(0, 10, N) + 7
    >>> y2 = np.sin(np.random.uniform(0, np.pi, N))
    >>> y3 = np.random.binomial(300, 1 / 10, N)
    >>> y4 = np.random.binomial(200, 1 / 3, N)
    >>> y5 = np.random.uniform(0, 800, N)
    ... # combine all arrays into a pandas DataFrame
    >>> data_np = np.column_stack((y1, y2, y3, y4, y5))
    >>> data_df = pd.DataFrame(data_np, columns=ynames)
    ... # using a DataFrame to draw parallel coordinates
    >>> parallel_coordinates(data_df, names=ynames)
    ... # using continuous values for categories
    >>> parallel_coordinates(data_df, names=ynames, categories=np.random.randint(0, 5, N))
    ... # using categorical classes
    >>> parallel_coordinates(data_df, names=ynames, categories=random.choices(categories_, k=N))
    ... # using numpy array instead of DataFrame
    >>> parallel_coordinates(data_df.values, names=ynames)
    ... # with customized tick labels
    >>> parallel_coordinates(data_df.values, ticklabel_kws={"fontsize": 8, "color": "red"})
    ... # using straight lines instead of bezier
    >>> parallel_coordinates(data_df, linestyle="straight")
    ... # with categorical class labels
    >>> data_df['P5'] = random.choices(categories_, k=N)
    >>> parallel_coordinates(data_df, names=ynames)
    ... # with categorical class labels and customized ticklabels
    >>> data_df['P5'] = random.choices(categories_, k=N)
    >>> parallel_coordinates(data_df,  ticklabel_kws={"fontsize": 8, "color": "red"})

    See :ref:`sphx_glr_auto_examples_parallel_coordinates.py` for more examples

    Note
    ----
        If nans are present in data or categories, all the corresponding enteries/rows
        will be removed.
    """

    try:
        import pandas as pd
    except (ModuleNotFoundError, ImportError):
        raise NotImplemented(f"You must install pandas to draw parallel plot")

    if cmap is None:
        cmap = "Blues"

    if isinstance(data, np.ndarray):
        assert data.ndim == 2, f"{data.ndim} dimensional data not allowed. It must be 2d"
        if names is None:
            names = [f"Feat_{i}" for i in range(data.shape[1])]
        data = pd.DataFrame(data, columns=names)

    if hasattr(data, "columns"):
        names = names or data.columns.tolist()

    if len(names) != data.shape[1]:
        raise ValueError(f"""
            provided names have length {len(names)} but data has {data.shape[1]} columns""")

    show_colorbar = True
    if categories is None:
        show_colorbar = False
        categories = np.linspace(0, 1, len(data))

    categories = np.array(categories)
    assert len(categories) == len(data)

    # remove NaN values based upon nan values in data
    if data.isna().sum().sum() > 0:
        df_nan_idx = data.isna().any(axis=1)
        categories = categories[~df_nan_idx]
        data = data[~df_nan_idx]

    _is_categorical = False
    cat_encoded = categories
    if not np.issubdtype(categories.dtype, np.number):
        # category contains categorical/non-numeri values
        cat_encoded = label_encoder(categories)
        _is_categorical = True

    if not _is_categorical:  # because we can't do np.isnan for categorical values
        # if there are still any nans in categories, remove them
        cat_nan_idx = np.isnan(categories)
        if cat_nan_idx.any():
            categories = categories[~cat_nan_idx]
            data = data[~cat_nan_idx]

    num_cols = data.shape[1]
    num_lines = len(data)

    # find out which columns are categorical and which are numerical
    enc_data = data.copy()
    cols = {}
    for idx, col in enumerate(data.columns):
        _col = data[col]
        if is_categorical(data[col].values):
            col_encoded = label_encoder(data[col].values)
            cols[idx] = {'cat': True, 'original': _col}
            enc_data[col] = col_encoded
        else:
            cols[idx] = {'cat': False}

    # organize the data
    enc_data = enc_data.astype(float)
    ymins = np.min(enc_data.values, axis=0)  # ys.min(axis=0)
    ymaxs = np.max(enc_data.values, axis=0)  # ys.max(axis=0)
    dys = ymaxs - ymins
    ymins -= dys * 0.05  # add 5% padding below and above
    ymaxs += dys * 0.05
    dys = ymaxs - ymins

    # transform all data to be compatible with the main axis
    zs = np.zeros_like(enc_data.values)
    zs[:, 0] = enc_data.iloc[:, 0]
    zs[:, 1:] = (enc_data.iloc[:, 1:] - ymins[1:]) / dys[1:] * dys[0] + ymins[0]

    fig, host = plt.subplots(figsize=figsize)

    axes = [host] + [host.twinx() for _ in range(num_cols - 1)]
    for i, ax in enumerate(axes):
        ax.set_ylim(ymins[i], ymaxs[i])
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        if ax != host:
            ax.spines['left'].set_visible(False)
            ax.yaxis.set_ticks_position('right')
            ax.spines["right"].set_position(("axes", i / (num_cols - 1)))

        if cols[i]['cat']:
            categories = np.unique(cols[i]['original'])
            new_ticks = np.unique(enc_data.iloc[:, i]).astype("float32")
            ax.set_yticks(new_ticks)
            ax.set_yticklabels(categories)

        if ticklabel_kws:
            if cols[i]['cat']:
                ticks_loc = [l._text for l in ax.get_yticklabels()]
            else:
                ticks_loc = ax.get_yticks().tolist()

            ax.set_yticks(ax.get_yticks().tolist())
            ax.set_yticklabels([label_format(x) for x in ticks_loc], **ticklabel_kws)

    if coord_title_kws is None:
        coord_title_kws = dict()

    _coord_title_kws = {'fontsize': 14}

    _coord_title_kws.update(coord_title_kws)

    host.set_xlim(0, num_cols - 1)
    host.set_xticks(range(num_cols))
    host.set_xticklabels(names, **_coord_title_kws)
    host.tick_params(axis='x', which='major', pad=7)
    host.spines['right'].set_visible(False)
    host.xaxis.tick_top()
    if title:
        host.set_title(title, fontsize=18)

    # category between 0.2,1 to map colors to their values
    cat_norm = _rescale(cat_encoded, 0.2)

    for j in range(num_lines):
        # color of each line is based upon corresponding value in category
        colors = getattr(cm, cmap)(cat_norm[j])

        if linestyle == "straight":
            # to just draw straight lines between the axes:
            host.plot(range(num_cols), zs[j, :], c=colors)
        else:
            # create bezier curves
            # for each axis, there will a control vertex at the point itself, one at 1/3rd towards the previous and one
            #   at one third towards the next axis; the first and last axis have one less control vertex
            # x-coordinate of the control vertices: at each integer (for the axes) and two inbetween
            # y-coordinate: repeat every point three times, except the first and last only twice
            x_coords = [x for x in np.linspace(0, len(data) - 1, len(data) * 3 - 2, endpoint=True)]
            y_coords = np.repeat(zs[j, :], 3)[1:-1]
            verts = list(zip(x_coords, y_coords))
            # for x,y in verts: host.plot(x, y, 'go') # to show the control points of the beziers
            codes = [Path.MOVETO] + [Path.CURVE4 for _ in range(len(verts) - 1)]
            path = Path(verts, codes)
            patch = patches.PathPatch(path, facecolor='none', lw=1, edgecolor=colors)
            host.add_patch(patch)

    if show_colorbar:
        norm = cm.colors.Normalize(np.min(cat_encoded), np.max(cat_encoded))

        cb = cm.ScalarMappable(norm, cmap=cmap)

        if _is_categorical:
            cbar = fig.colorbar(cb, orientation="vertical", pad=0.1)
            ticks = cbar.get_ticks()
            new_ticks = np.linspace(ticks[0], ticks[-1], len(np.unique(categories)))
            cbar.set_ticks(new_ticks)
            cbar.set_ticklabels(np.unique(categories))

        else:
            cbar = fig.colorbar(cb, orientation="vertical", pad=0.1)

    plt.tight_layout()

    if show:
        plt.show()

    return host


def label_format(x):
    if isinstance(x, float):
        return round(x, 3)
    else:
        return x


def is_categorical(array) -> bool:
    return not np.issubdtype(array.dtype, np.number)


def label_encoder(arr):
    # label encoder of numpy array with categorical values
    return np.unique(arr, return_inverse=True)[1]




__all__ = ["surf"]

from typing import Union, Tuple

import numpy as np
import matplotlib.pyplot as plt

def surf(
        x,
        y,
        z: Union[np.ndarray, Tuple[np.ndarray]] = None,
        c: Union[np.ndarray, Tuple[np.ndarray]] = None,
        facecolor = None,
        edgecolor = None,
        alpha = None,
        ax=None,
        show=True,
)->plt.Axes:
    """
    Plots surface. The purpose is to ease the drawing of surfaces
    using matplotlib and replicate matlab's surf function.

    parameters
    ----------
    x :
    y :
    z :
    c : Optional (default=None)
        array defining color of surface
    facecolor :
    edgecolor :
    alpha :
    ax : plt.Axes, optional (default=None)
        the matplotlib Axes object on which to plot the surface.
        If nto given, current available axes is taken using plt.gca()
    show : bool, optional (default=None)
        whether to show the plot or not.

    Returns
    --------
    plt.Axes
        matplotlib axes object on which surface is drawn

    Examples
    --------
    >>> from easy_mpl import surf
    >>> X = np.random.random(100)
    >>> Y = np.random.random(100)
    >> surf(X, Y)
    >>> Z = np.random.random(100)
    >>> surf(X, Y, Z)
    >>> C = np.random.random(100)
    >>> surf(X, Y, Z, C)
    """
    if not ax:
        ax = plt.gca()

    if show:
        plt.show()

    raise NotImplementedError




__all__ = ['taylor_plot']

from typing import Union
import pprint

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.projections import PolarAxes
import mpl_toolkits.axisartist.grid_finder as GF
import mpl_toolkits.axisartist.floating_axes as FA


COLORS = {
    0: np.array([0.89411765, 0.10196078, 0.10980392, 1.]),
    1: np.array([0.21568627, 0.49411765, 0.72156863, 1.]),
    2: np.array([0.30196078, 0.68627451, 0.29019608, 1.]),
    3: np.array([0.59607843, 0.30588235, 0.63921569, 1.]),
    4: np.array([0.95896577, 0.58394066, 0.04189788, 1.]),
    5: np.array([0.96862745, 0.50588235, 0.74901961, 1.]),
    6: np.array([0.17877267, 0.78893675, 0.92613355, 1.]),
    7: np.array([0.38079258, 0.17830983, 0.78165943, 1.]),
    8: np.array([0.13778617, 0.06228198, 0.33547859, 1.]),
    9: np.array([0.96707953, 0.46268314, 0.45772886, 1.]),
    10: np.array([0.17221373, 0.53023578, 0.96788307, 1.]),
    11: np.array([0.92875036, 0.09364162, 0.33348078, 1.]),
    12: np.array([0.93950089, 0.64582256, 0.16928645, 1.]),
    13: np.array([0.06802773, 0.46382623, 0.49007703, 1.]),
    14: np.array([0.13684922, 0.98802401, 0.34518303, 1.]),
    15: np.array(np.random.random(3).tolist() + [1.]),
    16: np.array([0.54829269, 0.15069842, 0.06147751, 1.]),
    17: np.array([0.61, 0.59, 0.51, 1.]),
    18: np.array([0.54829269, 0.15069842, 0.06147751, 1.]),
    19: np.array([0.06802773, 0.46382623, 0.49007703, 1.]),
    20: np.array([0.80, 0.50, 0.59, 1.]),
    21: np.array(np.random.random(3).tolist() + [1.]),
    22: np.array(np.random.random(3).tolist() + [1.]),
    23: np.array(np.random.random(3).tolist() + [1.]),
    24: np.array(np.random.random(3).tolist() + [1.]),
    25: np.array(np.random.random(3).tolist() + [1.]),
    26: np.array(np.random.random(3).tolist() + [1.]),
    27: np.array(np.random.random(3).tolist() + [1.]),
    28: np.array(np.random.random(3).tolist() + [1.]),
    29: np.array(np.random.random(3).tolist() + [1.]),
    30: np.array(np.random.random(3).tolist() + [1.]),
    31: np.array(np.random.random(3).tolist() + [1.]),
    32: np.array(np.random.random(3).tolist() + [1.]),
    33: np.array(np.random.random(3).tolist() + [1.]),
    34: np.array(np.random.random(3).tolist() + [1.]),
    35: np.array(np.random.random(3).tolist() + [1.]),
    36: np.array(np.random.random(3).tolist() + [1.]),
    37: np.array(np.random.random(3).tolist() + [1.]),
    38: np.array(np.random.random(3).tolist() + [1.]),
    39: np.array(np.random.random(3).tolist() + [1.]),
    40: np.array(np.random.random(3).tolist() + [1.]),
    41: np.array(np.random.random(3).tolist() + [1.]),
    42: np.array(np.random.random(3).tolist() + [1.]),
    43: np.array(np.random.random(3).tolist() + [1.]),
    44: np.array(np.random.random(3).tolist() + [1.]),
    45: np.array(np.random.random(3).tolist() + [1.]),
    46: np.array(np.random.random(3).tolist() + [1.]),
    47: np.array(np.random.random(3).tolist() + [1.]),
    48: np.array(np.random.random(3).tolist() + [1.]),
    49: np.array(np.random.random(3).tolist() + [1.]),
    50: np.array(np.random.random(3).tolist() + [1.]),
    51: np.array(np.random.random(3).tolist() + [1.]),
    52: np.array(np.random.random(3).tolist() + [1.]),
}

RECTS = {1: (111,),
         2: (211, 212),
         3: (311, 312, 313),
         4: (221, 222, 223, 224)}


def corr_coeff(t, p)->float:
    """correlation coefficient"""
    return np.corrcoef(t,p)[0,1]


def pbias(t, p)->float:
    """percent bias"""
    return np.mean(t - p) / np.mean(p)




class TaylorDiagram(object):
    """
    Taylor diagram.
    Modified after https://gist.github.com/ycopin/3342888
    Plot model standard deviation and correlation to reference (data)
    sample in a single-quadrant polar plot, with r=stddev and
    theta=arccos(correlation).
    """

    def __init__(self,
                 true,
                 fig=None,
                 rect=111,
                 label='_',
                 srange=(0, 1.5),
                 extend=False,
                 axis_fontdict: dict = None,
                 name:str = ''
                 ):
        """
        Set up Taylor diagram axes, i.e. single quadrant polar
        plot, using `mpl_toolkits.axisartist.floating_axes`.
        Parameters:
            true : true array or a dictionary containing standard deviation of the true
            fig : input Figure or None
            rect : subplot definition
            label : reference label
            srange : stddev axis extension, in units of *refstd*
            extend : extend diagram to negative correlations
            axis_fontdict : dictionary must consist of at least three dictionaries 'left', 'right', 'top'
        """

        # Reference standard deviation
        if isinstance(true, dict):
            self.refstd = true['std']
        else:
            self.refstd = np.std(true)

        if hasattr(self.refstd, 'item'):  # if true is series/dataframe
            self.refstd = self.refstd.item()

        assert not np.isnan(self.refstd), f"""
        stdandard deviation of {name} Observations is nan."""

        tr = PolarAxes.PolarTransform()

        # Correlation labels
        rlocs = np.array([0, 0.2, 0.4, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99, 1])
        if extend:
            # Diagram extended to negative correlations
            self.tmax = np.pi
            rlocs = np.concatenate((-rlocs[:0:-1], rlocs))
        else:
            # Diagram limited to positive correlations
            self.tmax = np.pi/2
        tlocs = np.arccos(rlocs)        # Conversion to polar angles
        gl1 = GF.FixedLocator(tlocs)    # Positions
        tf1 = GF.DictFormatter(dict(zip(tlocs, map(str, rlocs))))

        # Standard deviation axis extent (in units of reference stddev)
        self.smin = srange[0] * self.refstd
        self.smax = srange[1] * self.refstd

        ghelper = FA.GridHelperCurveLinear(
            tr,
            extremes=(0, self.tmax, self.smin, self.smax),
            grid_locator1=gl1, tick_formatter1=tf1)

        if fig is None:
            fig = plt.figure()

        ax = FA.FloatingSubplot(fig, rect, grid_helper=ghelper)
        fig.add_subplot(ax)

        # Adjust axes
        ax.axis["top"].set_axis_direction("bottom")   # "Angle axis"
        ax.axis["top"].toggle(ticklabels=True, label=True)
        ax.axis["top"].major_ticklabels.set_axis_direction("top")
        ax.axis["top"].label.set_axis_direction("top")
        ax.axis["top"].label.set_text("Correlation")
        ax.axis['top'].label.set_fontsize(axis_fontdict['top'].get('fontsize', 18))
        ax.axis['top'].label.set_color(axis_fontdict['top'].get('color', 'k'))
        ax.axis['top'].major_ticklabels.set_fontsize(axis_fontdict['top'].get('ticklabel_fs', 10))

        ax.axis["left"].set_axis_direction("bottom")  # "X axis"
        ax.axis["left"].label.set_text("Standard deviation")
        ax.axis['left'].label.set_fontsize(axis_fontdict['bottom'].get('fontsize', 18))
        ax.axis['left'].label.set_color(axis_fontdict['bottom'].get('color', 'k'))
        ax.axis['left'].major_ticklabels.set_fontsize(axis_fontdict['bottom'].get('ticklabel_fs', 10))

        ax.axis["right"].set_axis_direction("top")    # "Y-axis"
        ax.axis["right"].toggle(ticklabels=True)
        ax.axis["right"].label.set_text("Standard deviation")  # todo, not working
        ax.axis['right'].label.set_fontsize(axis_fontdict['left'].get('fontsize', 18))
        ax.axis['right'].label.set_color(axis_fontdict['left'].get('color', 'k'))
        ax.axis['right'].major_ticklabels.set_fontsize(axis_fontdict['left'].get('ticklabel_fs', 10))
        ax.axis["right"].major_ticklabels.set_axis_direction(
            "bottom" if extend else "left")

        if self.smin:
            ax.axis["bottom"].toggle(ticklabels=False, label=False)
        else:
            ax.axis["bottom"].set_visible(False)          # Unused

        self._ax = ax                   # Graphical axes
        self.ax = ax.get_aux_axes(tr)   # Polar coordinates

        # Add reference point and stddev contour
        l, = self.ax.plot([0], self.refstd, 'k*',
                          ls='', ms=10, label=label)
        t = np.linspace(0, self.tmax)
        r = np.zeros_like(t) + self.refstd
        self.ax.plot(t, r, 'k--', label='_')

        # Collect sample points for latter use (e.g. legend)
        self.samplePoints = [l]

    def add_sample(self, stddev, corrcoef, *args, **kwargs):
        """
        Add sample (*stddev*, *corrcoeff*) to the Taylor
        diagram. *args* and *kwargs* are directly propagated to the
        `Figure.plot` command.
        """

        l, = self.ax.plot(np.arccos(corrcoef), stddev,
                          *args, **kwargs)  # (theta, radius)
        self.samplePoints.append(l)

        return l

    def add_grid(self, *args, **kwargs):
        """Add a grid."""

        self._ax.grid(*args, **kwargs)

    def add_contours(self, levels=5, **kwargs):
        """
        Add constant centered RMS difference contours, defined by *levels*.
        """

        rs, ts = np.meshgrid(np.linspace(self.smin, self.smax),
                             np.linspace(0, self.tmax))
        # Compute centered RMS difference
        rms = np.sqrt(self.refstd**2 + rs**2 - 2*self.refstd*rs*np.cos(ts))

        contours = self.ax.contour(ts, rs, rms, levels, **kwargs)

        return contours


def taylor_plot(
        observations: Union[dict, np.ndarray],
        simulations: Union[dict, np.ndarray],
        axis_locs: dict = None,
        cont_kws: dict = None,
        grid_kws: dict = None,
        leg_kws: dict = None,
        axis_fontdict: dict = None,
        axis_kws: dict = None,
        marker_kws : dict = None,
        **kwargs
) -> plt.Figure:
    """
    Helper function to plot `Taylor`_'s plot. A taylor plot is useful to compare
    results of different simulations.

    Arguments:
        observations : dict, array
            If it is an array then it means we have only one scenario. If it is
            a dictionary of length > 1, then its keys are scenarios and values
            represent true/observations at that scenarios. The values can also
            be a dictionary containing `std`, which stands for standard deviation.
        simulations : dict
            A dictionary of length > 1 whose keys are scenarios and whose values
            are also dictionary. Each sub-dictionary i.e. dictionary of scenario
            consist of models/simulations.
        axis_locs : dict, optional
            dictionary defining axis orientation of figure. For example with two
            scenarios named 'scenario1' and 'scenario2', if we want to plot two
            plots in one column, then this argument will be

            >>>          {'scenario1': 211,
            >>>           'scenario2': 212}.

            Default is None.
        cont_kws : dict, optional
            keyword arguments related to `contours`_. Following args can be used:

                - levels level of contours
                - colors color of contours
                - label_fs fontsize of labels
                - label_fmt format of labels
                - linewidths float or sequence of floats
                - linestyles {None, 'solid', 'dashed', 'dashdot', 'dotted'}

        grid_kws : dict, optional
            keyword arguments related to grid. Following args can be used.
            Following keyword arguments are allowed:

                - title_fontsize: int, fontsize of the axis title
                - which {'major', 'minor', 'both'}
                - axis {'both', 'x', 'y'},
            any kwargs from `axes.grid`_ are allowed.

        leg_kws : dict, optional
            keyword arguments related to legends:

                - position defaults to `center`
                - fontsize int or {'xx-small', 'x-small', 'small', 'medium', 'large',
                                   'x-large', 'xx-large'}
                - numpoints int, default: rcParams["legend.numpoints"] (default: 1)
                - markerscale float, default: rcParams["legend.markerscale"] (default: 1.0)
            see `axes.legend`_ for more details.

            >>> example leg_kws = {'loc': 'upper right', 'numpoints': 1, 'fontsize': 15,
            ...                    'markerscale': 1}

        axis_fontdict : dict, optional
            dictionary defining propertiies of left, bottom and top axis labels

            >>> axis_fontdict = {'left': {'fontsize': 20, 'color': 'k', 'ticklabel_fs': 14},
            ...                 'bottom': {'fontsize': 20, 'color': 'g', 'ticklabel_fs': 14},
            ...                 'top': {'fontsize': 20, 'color': 'k', 'ticklabel_fs': 14}}

            The user can define properties of either one or all axis.

        axis_kws : dict, optional
            dictionary containing general parameters related to axis such as title.
        
        marker_kws : dict, optional
            dictionary containing marker properties. All of these keywords are passed to
            `matplotlib.axes.Axes.plot`. 

        **kwargs : 
            Following keyword arguments are optional:

            - add_ith_interval : bool

            - plot_bias : bool, if True, the size of the markers will be used to
                represent bias. The markers will be triangles with their sides up/down
                depending upon value of bias.

            - ref_color : str, color of refrence dot

            - sim_marker : marker to use for simulations. It can be any valid
                marker for matplotlib axis/plot. If None, then counting is used.
                If string, then same marker is used for all simulations. If dict,
                keys of dict should match with names of models in `simulations` dictionary.

            - true_label : label to use for `trues`. Default is 'Reference'.

            - intervals : list, if add_ith_interval is True, then this argument is used. It
                must be list of lists or list of tuples, where the inner tuple/list must
                consist of two values one each for x and y.

            - colors : 2d numpy array, defining colors. The first dimension
                should be equal to number of models.

            - extend : bool, default False, if True, will plot negative correlation

            - figsize : tuple defining figsize, default is (11,8).

            - show : bool whether to show the plot or not
    Returns:
        matplotlib figure

    Example:
        >>> import numpy as np
        >>> from easy_mpl import taylor_plot
        >>> np.random.seed(313)
        >>> observations =  np.random.normal(20, 40, 10)
        >>> simulations =  {"LSTM": np.random.normal(20, 40, 10),
        ...             "CNN": np.random.normal(20, 40, 10),
        ...             "TCN": np.random.normal(20, 40, 10),
        ...             "CNN-LSTM": np.random.normal(20, 40, 10)}
        >>> taylor_plot(observations=observations,
        ...             simulations=simulations,
        ...             title="Taylor Plot")
        ... 
        ... # multiple taylor plots in one figure
        >>> np.random.seed(313)
        >>> observations = {
        ...     'site1': np.random.normal(20, 40, 10),
        ...     'site2': np.random.normal(20, 40, 10),
        ...     'site3': np.random.normal(20, 40, 10),
        ...     'site4': np.random.normal(20, 40, 10),
        ... }
        ... 
        >>> simulations = {
        ...     "site1": {"LSTM": np.random.normal(20, 40, 10),
        ...                 "CNN": np.random.normal(20, 40, 10),
        ...                 "TCN": np.random.normal(20, 40, 10),
        ...                 "CNN-LSTM": np.random.normal(20, 40, 10)},
        ... 
        ...     "site2": {"LSTM": np.random.normal(20, 40, 10),
        ...                 "CNN": np.random.normal(20, 40, 10),
        ...                 "TCN": np.random.normal(20, 40, 10),
        ...                 "CNN-LSTM": np.random.normal(20, 40, 10)},
        ... 
        ...     "site3": {"LSTM": np.random.normal(20, 40, 10),
        ...                 "CNN": np.random.normal(20, 40, 10),
        ...                 "TCN": np.random.normal(20, 40, 10),
        ...                 "CNN-LSTM": np.random.normal(20, 40, 10)},
        ... 
        ...     "site4": {"LSTM": np.random.normal(20, 40, 10),
        ...                 "CNN": np.random.normal(20, 40, 10),
        ...                 "TCN": np.random.normal(20, 40, 10),
        ...                 "CNN-LSTM": np.random.normal(20, 40, 10)},
        ... }
        ... # define positions of subplots
        >>> rects = dict(site1=221, site2=222, site3=223, site4=224)
        ... 
        >>> taylor_plot(observations=observations,
        ...             simulations=simulations,
        ...             axis_locs=rects,
        ...             plot_bias=True,
        ...             cont_kws={'colors': 'blue', 'linewidths': 1.0, 'linestyles': 'dotted'},
        ...             grid_kws={'axis': 'x', 'color': 'g', 'lw': 1.0},
        ...             title="mutiple subplots")
        ...

        Sometimes we don't have actual observation and simulation values as arrays. We can
        still make Taylor plot by providing only standard deviation of observations
        and simulations and coefficient of correlation (R) values
        between simulations and observations.

        >>> observations = {'std': 4.916}
        >>> predictions = {   # pbias is optional
        ...         'Model 1': {'std': 2.80068, 'corr_coeff': 0.49172, 'pbias': -8.85},
        ...         'Model 2': {'std': 3.47, 'corr_coeff': 0.67, 'pbias': -19.76},
        ...         'Model 3': {'std': 3.53, 'corr_coeff': 0.596, 'pbias': 7.81},
        ...         'Model 4': {'std': 2.36, 'corr_coeff': 0.27, 'pbias': -22.78},
        ...         'Model 5': {'std': 2.97, 'corr_coeff': 0.452, 'pbias': -7.99}}
        ... 
        >>> taylor_plot(observations,
        ...     predictions,
        ...     title="with statistical parameters")
        ... 
        ... # with customized markers
        >>> np.random.seed(313)
        >>> observations =  np.random.normal(20, 40, 10)
        >>> simulations =  {"LSTM": np.random.normal(20, 40, 10),
        ...                 "CNN": np.random.normal(20, 40, 10),
        ...                 "TCN": np.random.normal(20, 40, 10),
        ...                 "CNN-LSTM": np.random.normal(20, 40, 10)}
        >>> taylor_plot(observations=observations,
        ...             simulations=simulations,
        ...             title="customized markers",
        ...             marker_kws={'markersize': 10, 'markeredgewidth': 1.5,
        ...                 'markeredgecolor': 'black'})
        ... 
        ... # with customizing bbox
        >>> np.random.seed(313)
        >>> observations =  np.random.normal(20, 40, 10)
        >>> simus =  {"LSTMBasedRegressionModel": np.random.normal(20, 40, 10),
        ...         "CNNBasedRegressionModel": np.random.normal(20, 40, 10),
        ...         "TCNBasedRegressionModel": np.random.normal(20, 40, 10),
        ...         "CNN-LSTMBasedRegressionModel": np.random.normal(20, 40, 10)}
        >>> taylor_plot(observations=observations,
        ...             simulations=simus,
        ...             title="custom_legend",
        ...             leg_kws={'facecolor': 'white',
        ...                 'edgecolor': 'black','bbox_to_anchor':(1.1, 1.05)})

    See :ref:`sphx_glr_auto_examples_taylor_plot.py` for more examples

    .. _Taylor:
        https://doi.org/10.1029/2000JD900719 

    .. _contours:
        https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.contour.html
    
    .. _axes.grid:
        https://matplotlib.org/3.3.3/api/_as_gen/matplotlib.axes.Axes.grid.html
    
    .. _axes.legend:
        https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.legend.html

    """
    # todo replicate https://bookdown.org/david_carslaw/openair/sec-TaylorDiagram.html
    values = False

    if isinstance(observations, np.ndarray): 
        observations = {'Reference': observations}
        simulations = {'Reference': simulations}

    if isinstance(observations, dict) and len(observations) == 1 and 'std' in observations.keys():
        observations = {'Reference': observations}
        simulations = {'Reference': simulations}

    if all([isinstance(v, dict) for v in observations.values()]):  # we are given std, corr_coeff values
        values = True
    else:  # we are given actual arrays
        assert all([isinstance(np.array(v), np.ndarray) for v in observations.values()])
        assert all([len(array) > 1 for array in observations.values()]), f"""
        one or more array in true values has less than 2 values
        {pprint.pprint({key: len(array) for key, array in observations.items()}, width=20)}"""
    scenarios = observations.keys()

    figsizes = {
        1: (8, 6),
        2: (11, 9),
        3: (11, 9),
        4: (12, 10)
    }

    n_plots = len(observations)
    assert n_plots == len(simulations)

    add_ith_interval = kwargs.get('add_idth_interval', False)
    ref_color = kwargs.get('ref_color', 'r')
    intervals = kwargs.get('intervals', [])
    colors = kwargs.get('colors', COLORS)
    extend = kwargs.get('extend', False)
    plot_bias = kwargs.get('plot_bias', False)
    title = kwargs.get('title', "")
    figsize = kwargs.get("figsize", figsizes[n_plots])  # widht and heigt respectively
    sim_marker = kwargs.get("sim_marker", None)
    true_label = kwargs.get("true_label", "Reference")
    show = kwargs.get("show", True)

    if axis_locs is None:
        axis_locs = {k: v for k, v in zip(scenarios, RECTS[len(scenarios)])}

    # importing it at the start will be too early, because every sub-module imports from utils
    # so utils must not import something from other sub-modules at the start
    

    sims = list(simulations.values())
    models = len(sims[0])

    for m in sims:
        assert len(m) == models

    def msg(key, where="simulations"):
        return f"Scenario {key} does not match any of the provided scenarios in {where}"

    for scen in scenarios:
        if scen not in simulations:
            raise KeyError(msg(scen))
        if scen not in axis_locs:
            raise KeyError(msg(scen, "axis_locs"))

    def get_marker(third_value, _name):
        marker = 'o'

        if sim_marker is not None:
            if isinstance(sim_marker, str):
                return sim_marker
            elif isinstance(sim_marker, dict):
                return sim_marker[_name]

        if third_value:

            if third_value >= 0.0:
                marker = "^"
            else:
                marker = "v"

        return marker
    
    def get_marker_kws(index):
        if marker_kws:
            return marker_kws

        return {'ms':10, 'ls':'', 'mfc': colors[index], 'mec': colors[index], 'mew':1.0} 

    plt.close('all')
    fig = plt.figure(figsize=figsize)
    if title is not None:
        fig.suptitle(title, fontsize=18)

    if axis_fontdict is None:
        axis_fontdict = {'left': {},
                         'right': {},
                         'bottom': {},
                         'top': {}}
    else:
        assert isinstance(axis_fontdict, dict)
        for k in ['left', 'right', 'bottom', 'top']:
            if k not in axis_fontdict:
                axis_fontdict[k] = {}

    add_grid = True
    grid_kws = grid_kws or {}

    if grid_kws is False:
        add_grid = False

    for season in scenarios:

        dia = TaylorDiagram(observations[season],
                            fig=fig,
                            rect=axis_locs[season],
                            label=true_label,
                            axis_fontdict=axis_fontdict,
                            extend=extend,
                            name=season,
                            )

        dia.samplePoints[0].set_color(ref_color)  # Mark reference point as a red star

        if add_ith_interval:
            for interval in intervals:
                dia.ax.plot(*interval, color='k')

        # Add samples to Taylor diagram
        idx = 0
        for model_name, model in simulations[season].items():
            if values:
                assert isinstance(model, dict)
                stddev = model['std']
                corrcoef = model['corr_coeff']
                third_val = None

                if len(model) > 2:
                    _ = list(model.keys())[2]
                    third_val = list(model.values())[2]
            else:
                assert model.size == len(model)
                assert observations[season].size == len(observations[season])
                stddev = np.std(model)
                corrcoef = corr_coeff(observations[season].reshape(-1,), model.reshape(-1,))
                third_val = None
                if plot_bias:
                    third_val = pbias(observations[season], model)

            marker = get_marker(third_val, model_name)

            dia.add_sample(stddev, corrcoef,
                           marker=marker,
                           **get_marker_kws(idx),
                           label=model_name)

            idx += 1

        if cont_kws is None:
            cont_kws = dict()

        # Add RMS contours, and label them
        contours = dia.add_contours(levels=cont_kws.get('level', 5), colors=cont_kws.get('colors', '0.5'),
                                    linewidths=cont_kws.get('linewidths', 1.5),
                                    linestyles=cont_kws.get('linestyles', None))  # 5 levels
        dia.ax.clabel(contours, inline=cont_kws.get('inline', 1), fontsize=cont_kws.get('label_fs', 10), fmt='%.1f')

        if add_grid:
            dia.add_grid(**grid_kws)  # Add grid
            dia._ax.axis[:].major_ticks.set_tick_out(True)  # Put ticks outward

        if axis_kws is None:
            axis_kws = dict()
        title_fontsize = axis_kws.pop('title_fontsize') if 'title_fontsize' in axis_kws else 14
        # Tricky: ax is the polar ax (used for plots), _ax is the
        # container (used for layout)
        if len(scenarios) > 1:
            dia._ax.set_title(season.capitalize(), fontdict={'fontsize': title_fontsize})

    # Add a figure legend and title. For loc option, place x,y tuple inside [ ].
    # Can also use special options here:
    # http://matplotlib.sourceforge.net/users/legend_guide.html

    if leg_kws is None:
        position = "center" if len(scenarios) == 4 else "upper right"
        leg_kws = {'loc': position}

    fig.legend(dia.samplePoints,
               [p.get_label() for p in dia.samplePoints],
               **leg_kws)

    fig.tight_layout()

    if show:
        plt.show()

    return fig




__all__ = ["contour"]

import numpy as np
import matplotlib.pyplot as plt

from .utils import process_axes


def contour(
        x,
        y,
        z,
        fill_between: bool = False,
        show_points: bool = False,
        colorbar: bool = True,
        label_contours: bool = False,
        contour_kws: dict = None,
        fill_between_kws: dict = None,
        show_points_kws: dict = None,
        label_contour_kws: dict = None,
        ax: plt.Axes = None,
        show: bool = True,
        ax_kws:dict = None,
):
    """A contour plot of irregularly spaced data coordinates.

    Parameters
    ---------
        x : array, list
            a 1d array defining grid positions along x-axis
        y : array, list
            a 1d array defining grid positions along y-axis
        z : array, list
            values on grids/points defined by x and y
        fill_between : bool, optional
            whether to fill the space between colors or not
        show_points : bool, optional
            whether to show the
        label_contours : bool, optional
            whether to label the contour lines or not
        colorbar : bool, optional
            whether to show the colorbar or not. Only valid if `fill_between` is
            True
        contour_kws : dict, optional
            any keword argument for `axes.tricontour`_
        fill_between_kws : dict, optional
            any keword argument for `axes.tricontourf`_
        show_points_kws : dict, optional
            any keword argument for `axes.plot`_
        label_contour_kws : dict, optional
            any keword argument for `axes.clabel`_
        ax : plt.Axes, optional
            matplotlib axes to work on. If not given, current active axes will
            be used.
        ax_kws : dict optional
            any keyword arguments for :py:func:`easy_mpl.utils.process_axes`.
        show : bool, optional
            whether to show the plot or not

    Returns
    -------
    plt.Axes
        a matplotliblib Axes

    Examples
    --------
        >>> from easy_mpl import contour
        >>> import numpy as np
        >>> _x = np.random.uniform(-2, 2, 200)
        >>> _y = np.random.uniform(-2, 2, 200)
        >>> _z = _x * np.exp(-_x**2 - _y**2)
        >>> contour(_x, _y, _z, fill_between=True, show_points=True)
        ... # show contour labels
        >>> contour(_x, _y, _z, label_contours=True, show_points=True)

    See :ref:`sphx_glr_auto_examples_contour.py` for more examples

    Note
    ----
    The length of x and y should be same. The actual grid is created using axes.tricontour
    and axes.tricontourf functions.

    .. _axes.tricontour:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.tricontour.html

    .. _axes.tricontourf:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.tricontourf.html

    .. _axes.plot:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html

    .. _axes.clabel:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.clabel.html
    """
    x = np.array(x)
    y = np.array(y)
    z = np.array(z)

    assert len(x) == len(y) == len(z)

    if ax_kws is None:
        ax_kws = dict()

    if ax is None:
        ax = plt.gca()
        if 'figsize' in ax_kws:
            figsize = ax_kws.pop('figsize')
            ax.figure.set_size_inches(figsize)

    kws = contour_kws or {"levels": 14, "linewidth": 0.5, "colors": "k"}
    CS = ax.tricontour(x, y, z, **kws)

    if fill_between:
        kws = fill_between_kws or {'levels': 14, 'cmap': 'RdBu_r'}
        CS = ax.tricontourf(x, y, z, **kws)

        if colorbar:
            fig: plt.Figure = plt.gcf()
            fig.colorbar(CS, ax=ax)

    if show_points:
        kws = show_points_kws or {'color': 'k', 'marker': 'o', 'ms': 3, "linestyle": ""}
        ax.plot(x, y, **kws)

    if ax_kws:
        process_axes(ax, **ax_kws)

    if label_contours:
        kws = label_contour_kws or {"colors": "black"}
        ax.clabel(CS, **kws)

    if show:
        plt.show()

    return ax




__all__ = ["scatter"]

from typing import Tuple, Union

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from .utils import to_1d_array, process_axes


def scatter(
        x,
        y,
        colorbar: bool = False,
        colorbar_orientation: str = "vertical",
        marker_labels: Union[list, np.ndarray] = None,
        text_kws : dict = None,
        xoffset = 0.1,
        yoffset = 0.1,
        ax: plt.Axes = None,
        ax_kws:dict = None,
        show: bool = True,
        **kwargs
) -> Tuple[plt.Axes, mpl.collections.PathCollection]:
    """
    scatter plot between two arrays x and y

    Parameters
    ----------
    x : list, array
        data for x-axis
    y : list, array
        data for y-axis
    colorbar : bool, optional
        whether to show the color bar or not
    colorbar_orientation : str, optional
        orientation of colorbar. Only relevant if ``colorbar`` is True
    marker_labels : list, array
        labels to annotate each marker. If given, each value must
        correspond to respective values in x,y arrays
    text_kws : dict
        only relevant if ``marker_labels`` are provided.
    xoffset : float
    yoffset : float
    ax : plt.Axes, optional
        :obj:`matplotlib.axes`, if not given, current available axes will be used
    ax_kws : dict (default=None)
        any keyword arguments for processing of axes which will
        be forwarded to :func:`easy_mpl.utils.prcess_axis`
    show : bool, optional (default=True)
        whether to show the plot or not
    **kwargs : optional
        any additional keyword arguments for :obj:`matplotlib.axes.Axes.scatter`

    Returns
    --------
    tuple :
        A tuple whose first member is :obj:`matplotlib.axes` and second member is
        :obj:`matplotlib.collections.PathCollection`

    Examples
    --------
        >>> from easy_mpl import scatter
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> x_ = np.random.random(100)
        >>> y_ = np.random.random(100)
        >>> scatter(x_, y_, show=False)
        ... # show colorbar
        >>> scatter(x_, y_, colorbar=True, show=False)
        ... # retrieve axes for further processing
        >>> axes, _ = scatter(x_, y_, show=False)
        >>> assert isinstance(axes, plt.Axes)

    See :ref:`sphx_glr_auto_examples_scatter.py` for more examples

    """
    if ax is None:
        ax = plt.gca()
        if 'figsize' in kwargs:
            figsize = kwargs.pop('figsize')
            ax.figure.set_size_inches(figsize)

    x = to_1d_array(x)
    y = to_1d_array(y)

    #if colorbar:
    #    if 'c' not in kwargs and 'color' not in kwargs:
    #        kwargs['c'] = np.arange(len(x))

    sc = ax.scatter(x, y, **kwargs)

    if marker_labels is not None:
        y = y.reshape(-1,)
        _text_kws = {}
        if text_kws is not None:
            _text_kws.update(text_kws)
        for i, txt in enumerate(marker_labels):
            ax.annotate(txt, (x[i] + xoffset, y[i] + yoffset), **_text_kws)

    if colorbar:
        fig: plt.Figure = ax.get_figure()
        fig.colorbar(sc, orientation=colorbar_orientation, pad=0.1)

    if ax_kws:
        process_axes(ax=ax, **ax_kws)

    if show:
        plt.show()

    return ax, sc




__all__ = ["circular_bar_plot"]

import random
import warnings
from typing import Union

import numpy as np
import matplotlib.pyplot as plt

from .utils import _rescale
from .utils import BAR_CMAPS
from .utils import to_1d_array, make_cols_from_cmap, process_axes
from .utils import map_array_to_cmap


def circular_bar_plot(
        data,
        labels: list = None,
        sort:bool =False,
        color: Union[str, list, np.ndarray] = None,
        label_format: str = None,
        min_max_range: tuple = None,
        label_padding: int = 4,
        figsize: tuple = None,
        text_kws: dict = None,
        colorbar:bool = False,
        ax:plt.Axes = None,
        ax_kws: dict = None,
        show: bool = True,
        **kwargs
) -> plt.Axes:
    """
    Plot a circular bar plot.

    Parameters
    ----------
    data : list, np.ndarray, pd.Series, dict
        Data to plot. If it is a dictionary, then its keys will be used
        as labels and values will be used as data.
    labels : list, optional
        Labels for each data point.
    sort : bool, optional
        Sort the data by the values.
    color : str, list, np.ndarray, optional
        Color for each data point. It can be a single color or a colormap from
        plt.colormaps.
    label_format : str, optional
        Format for the labels.
    min_max_range : tuple, optional
        Minimum and maximum range for normalizing the data.
    label_padding : int, optional
        space between the labels and the bars.
    figsize : tuple, optional
        Size of the figure.
    text_kws : dict, optional (default=None)
        keyword arguments for axes.text()
    colorbar : bool (default=False)
        whether to show the colorbar or not
    ax : plt.Axes (default=None)
        matplotlib axews with polar projections
    ax_kws : optional
        Additional keyword arguments to pass to the :py:func:`easy_mpl.utils.process_axes`.
    show : bool, optional (default=True)
        Show the plot.
    **kwargs
        go to :obj:`matplotlib.axes.Axes.bar`

    Returns
    -------
    :obj:`matplotlib.axes`
        Axes of the plot.

    See :ref:`sphx_glr_auto_examples_circular_bar_plot.py` for more examples

    Note
    ----
        If nan values are present in the data, they will be ignored.

    Examples
    --------
    >>> import numpy as np
    >>> from easy_mpl import circular_bar_plot
    >>> data = np.random.random(50, )
    ... # basic
    >>> circular_bar_plot(data)
    ... # with names
    >>> names = [f"{i}" for i in range(50)]
    >>> circular_bar_plot(data, names)
    ... # sort values
    >>> circular_bar_plot(data, names, sort=True)
    ... # custom color map
    >>> circular_bar_plot(data, names, color='viridis')
    ... # custom min and max range
    >>> circular_bar_plot(data, names, min_max_range=(1, 10), label_padding=1)
    ... # custom label format
    >>> circular_bar_plot(data, names, label_format='{} {:.4f}')

    """

    text_kws = text_kws or {}

    if ax is None:
        plt.close('all')
        plt.figure(figsize=figsize or (7, 10))
        ax = plt.subplot(111, polar=True)
    else:
        assert ax.name == "polar"

    plt.axis('off')

    if hasattr(data, "values") and hasattr(data, "columns"):
        values = data.values
    elif isinstance(data, dict):
        values = np.array(list(data.values()))
        labels = labels or list(data.keys())
    else:
        data = to_1d_array(data)
        values = data

    if labels is None:
        labels = ['' for _ in range(len(values))]
        label_format = label_format or "{} {:.2f}"
    else:
        label_format = label_format or "{}: {:.2f}"

    # remove nan values
    val_nan_idx = np.isnan(values)
    if val_nan_idx.any():
        values = values[~val_nan_idx]
        labels = [labels[i] for i in range(len(labels)) if not val_nan_idx[i]]

    mapper = None
    if color is None:
        #color = make_cols_from_cmap(random.choice(BAR_CMAPS), len(values), 0.2, 0.9)
        color, mapper = map_array_to_cmap(values, random.choice(BAR_CMAPS))
    elif isinstance(color, str) and color in plt.colormaps():
        #color = make_cols_from_cmap(color, len(values), 0.2)
        color, mapper = map_array_to_cmap(values, color)
    else:
        color = color

    assert len(values) == len(labels)

    min_max_range = min_max_range or (30, 100)
    lower_limit = min_max_range[0]
    heights = _rescale(values.reshape(-1, 1), lower_limit, min_max_range[1]).reshape(-1, )

    if sort:
        sort_idx = np.argsort(heights)
        heights = heights[sort_idx]
        labels = [labels[i] for i in sort_idx]
        values = values[sort_idx]

        if not isinstance(color, str):
            color = color[sort_idx]

    # Compute the width of each bar. In total we have 2*Pi = 360°
    width = 2 * np.pi / len(heights)

    # Compute the angle each bar is centered on:
    indexes = list(range(1, len(heights) + 1))
    angles = [element * width for element in indexes]

    _bar_kws = {'linewidth': 2, 'edgecolor': 'white', 'color': color}
    if kwargs:
        _bar_kws.update(kwargs)

    # Draw bars
    bars = ax.bar(
        x=angles,
        height=heights,
        width=width,
        bottom=lower_limit,
        **_bar_kws
    )

    # Add labels
    for bar, angle, label, val in zip(bars, angles, labels, values):

        label = label_format.format(label, val)

        # Labels are rotated. Rotation must be specified in degrees :(
        rotation = np.rad2deg(angle)

        # Flip some labels upside down
        if angle >= np.pi / 2 and angle < 3 * np.pi / 2:
            alignment = "right"
            rotation = rotation + 180
        else:
            alignment = "left"

        # Finally add the labels
        ax.text(
            x=angle,
            y=lower_limit + bar.get_height() + label_padding,
            s=label,
            ha=alignment,
            va='center',
            rotation=rotation,
            rotation_mode="anchor",
            **text_kws
        )

    if colorbar:
        if mapper:
            fig: plt.Figure = ax.get_figure()
            cbar = fig.colorbar(mapper, orientation='horizontal')
            cbar.ax.tick_params('both', labelsize=16)
        else:
            warnings.warn(f"For single color, colorbar does not make sense!")

    if ax_kws:
        process_axes(ax, **ax_kws)

    if show:
        plt.show()

    return ax




__all__ = ["lollipop_plot"]

import numpy as np
import matplotlib.pyplot as plt

from .utils import is_rgb
from .utils import to_1d_array, process_axes


def lollipop_plot(
        y,
        x=None,
        orientation: str = "vertical",
        sort: bool = False,
        line_style: str = '-',
        line_color: str = 'cyan',
        line_width: int = 1,
        line_kws: dict = None,
        marker_style: str = 'o',
        marker_color: str = 'teal',
        marker_size: int = 30,
        marker_kws: dict = None,
        ax: plt.Axes = None,
        ax_kws:dict = None,
        show: bool = True,
        **kwargs
) -> plt.Axes:
    """
    Plot a lollipop plot.

    Parameters
    ----------
    y : array_like, shape (n,), optional
        The y-coordinates of the data points.
    x : array_like, shape (n,)
        The x-coordinates of the data points.
    orientation : str, optional
        The orientation of the lollipops. Either "vertical" or "horizontal".
    sort : bool, optional
        Whether to sort the data points by their values or not.
        Only valid if `x` is not specified.
    line_style : str, optional
        The line style of the data points.
    line_color : str, optional
        The line color of the data points.
    line_width : float, optional
        The line width of the data points.
    line_kws : dict, optional
        The keyword arguments for the line. These arguments are passed to
        `matplotlib.axes.Axes.plot`_.
    marker_style : str, optional
        The marker style of the data points.
    marker_color : str, optional
        The marker color of the data points.
    marker_size : float, optional
        The marker size of the data points.
    marker_kws : dict, optional
        The keyword arguments for the marker. These arguments are passed to
        `matplotlib.axes.Axes.scatter`_.
    ax : matplotlib.axes.Axes, optional
        The axes to plot on. If not given, current available axes will be used.
    ax_kws : dict, optional
        any keyword arguments for :py:func:`easy_mpl.utils.process_axes`.
    show : bool, optional (default=True)
        whether to show the plot or not
    **kwargs : optional
        Additional keyword arguments to pass to the process_axis function.

    Returns
    -------
    plt.Axes :
        The axes on which the plot was drawn.

    Examples
    --------
    >>> import numpy as np
    >>> from easy_mpl import lollipop_plot
    >>> y = np.random.randint(0, 10, size=10)
    ... # vanilla lollipop plot
    >>> lollipop_plot(y, title="vanilla")
    ... # use both x and y
    >>> lollipop_plot(y, np.linspace(0, 100, len(y)), title="with x and y")
    ... # use custom line style
    >>> lollipop_plot(y, line_style='--', title="with custom linestyle")
    ... # use custom marker style
    >>> lollipop_plot(y, marker_style='D', title="with custom marker style")
    ... # sort the data points before plotting
    >>> lollipop_plot(y, sort=True, title="sort")
    ... # horzontal orientation of lollipops
    >>> y = np.random.randint(0, 20, size=10)
    >>> lollipop_plot(y, orientation="horizontal", title="horizontal")

    See :ref:`sphx_glr_auto_examples_lollipop_plot.py` for more examples

    .. _matplotlib.axes.Axes.plot:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.plot.html

    .. _matplotlib.axes.Axes.scatter:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.scatter.html
    """
    if ax_kws is None:
        ax_kws = dict()

    if ax is None:
        ax = plt.gca()
        if 'figsize' in ax_kws:
            figsize = ax_kws.pop('figsize')
            ax.figure.set_size_inches(figsize)

    y = to_1d_array(y)

    if sort:
        idx = np.argsort(y)
        y = y[idx]
        assert x is None

        if isinstance(line_color, (list, np.ndarray, tuple)):
            if is_rgb(line_color[0]) or isinstance(line_color[0], str):
                line_color = np.array(line_color)[idx]

    if x is None:
        x = np.arange(len(y))

    x = to_1d_array(x)

    marker_kws = marker_kws or {}
    line_kws = line_kws or {}

    if orientation == "vertical":
        _lollipop_vertical(ax, x, y, line_style, line_color, line_width, line_kws,
                           marker_style, marker_color, marker_size, marker_kws)
    else:
        _lollipop_horizontal(ax, x, y, line_style, line_color, line_width, line_kws,
                             marker_style, marker_color, marker_size, marker_kws)

    if ax_kws and kwargs:
        process_axes(ax=ax, **ax_kws, **kwargs)

    if show:
        plt.show()

    return ax


def _lollipop_vertical(ax, x, y, line_style, line_color, line_width, line_kws,
                       marker_style, marker_color, marker_size, marker_kws):

    if isinstance(marker_color, str) and marker_color in plt.colormaps():
        marker_kws['cmap'] = marker_color
    else:
        marker_kws['color'] = marker_color

    ax.scatter(x, y, marker=marker_style,
               s=marker_size, **marker_kws)

    if line_color in plt.colormaps():
        line_kws['cmap'] = line_color
    else:
        line_kws['color'] = line_color

    ax.vlines(x, np.zeros(len(x)), y,
              linestyle=line_style, linewidth=line_width, **line_kws)
    return ax


def _lollipop_horizontal(ax, x, y, line_style, line_color, line_width, line_kws,
                         marker_style, marker_color, marker_size, marker_kws):
    ax.scatter(y, x, marker=marker_style, color=marker_color,
               s=marker_size, **marker_kws)
    ax.hlines(x, np.zeros(len(y)), y, color=line_color,
              linestyle=line_style, linewidth=line_width, **line_kws)
    return ax





__all__ = ["imshow"]

from typing import Union

import numpy as np
import matplotlib.pyplot as plt

from .utils import despine_axes
from .utils import process_cbar
from .utils import process_axes
from .utils import annotate_imshow


def imshow(
        values,
        yticklabels=None,
        xticklabels=None,
        annotate:bool = False,
        annotate_kws:dict = None,
        colorbar: bool = False,
        grid_params: dict = None,
        mask : Union[bool, str, np.ndarray] = None,
        cbar_params: dict = None,
        ax:plt.Axes = None,
        ax_kws: dict = None,
        show:bool = True,
        **kwargs
):
    """
    One stop shop for matplotlib's imshow function

    Parameters
    ----------
        values: 2d array
            the image/data to show. It must bt 2 dimensional. It can also
            be dataframe.
        annotate : bool, optional
            whether to annotate the heatmap or not
        annotate_kws : dict, optional
            a dictionary with following possible keys

                - ha : horizontal alighnment (default="center")
                - va : vertical alighnment (default="center")
                - fmt : format (default='%.2f')
                - textcolors : colors for axes.text
                - threshold : threshold to be used for annotation
                - **kws : any other keyword argument for axes.text

        colorbar : bool, optional
            whether to draw colorbar or not
        xticklabels : list, optional
            tick labels for x-axis. For DataFrames, column names are used by default.
        yticklabels :  list, optional
            tick labels for y-axis. For DataFrames, index is used by default
        grid_params : dict, optional (default=None)
            parameters to process grid. Allowed keys in the dictionary are following
                - ``border``, bool
                - ``linestyle``
                - ``linewidth``
                - ``color``
        mask :
            This argument can be used to hide part of heatmap from being displayed.
                - True : will only show the lower half
                - ``upper`` will only show the lower half
                - ``lower`` will only show the upper half
        cbar_params : dict, optional
            params for colorbar. for example ``pad`` or ``orientation``
        ax : plt.Axes, optional
            if not given, current available axes will be used
        ax_kws : dict, optional (default=None)
            any keyword arguments for :py:func:`easy_mpl.utils.process_axes` function as dictionary
        show : bool, optional
            whether to show the plot or not
        **kwargs : optional
            any further keyword arguments for :obj:`matplotlib.axes.Axes.imshow`

    Returns
    -------
    matplotlib.image.AxesImage
        a :obj:`matplotlib.image.AxesImage`

    Examples
    --------
        >>> import numpy as np
        >>> from easy_mpl import imshow
        >>> x = np.random.random((10, 5))
        >>> imshow(x, annotate=True)
        ... # show colorbar
        >>> imshow(x, colorbar=True)
        ... # setting white grid lines and annotation
        >>> data = np.random.random((4, 10))
        >>> imshow(data, cmap="YlGn",
        ...        xticklabels=[f"Feature {i}" for i in range(data.shape[1])],
        ...        grid_params={'border': True, 'color': 'w', 'linewidth': 2}, annotate=True,
        ...        colorbar=True)

    See :ref:`sphx_glr_auto_examples_imshow.py` for more examples

    """
    if ax_kws is None:
        ax_kws = dict()

    if ax is None:
        ax = plt.gca()
        if 'figsize' in ax_kws:
            figsize = ax_kws.pop('figsize')
            ax.figure.set_size_inches(figsize)

    if hasattr(values, "values") and hasattr(values, "columns"):
        import pandas as pd  # don't make whole project dependent upon pandas

        if not xticklabels:
            xticklabels = values.columns.to_list()
        if not yticklabels:
            yticklabels = values.index.tolist()
        # when data in dataframe is object type, it causes error in plotting
        # the best way to convert series in df to number is to use to_numeric
        values = np.column_stack([pd.to_numeric(values.iloc[:, i]) for i in range(values.shape[1])])

    to_keep = None
    if mask is not None:
        if isinstance(mask, (str, bool)):
            _mask = np.tri(values.shape[0], k=-1)
            if mask == "lower":
                values = np.ma.array(values, mask=_mask)  # mask out the lower triangle
                to_keep = ['right', 'top']
            else:
                values = np.ma.array(values, mask=_mask).T
                to_keep = ['left', 'bottom']

    tick_params = {}
    if 'ticks' in kwargs:
        tick_params['ticks'] = kwargs.pop('ticks')

    im = ax.imshow(values, **kwargs)

    if to_keep:
        despine_axes(ax, keep=to_keep)


    if annotate_kws is None:
        annotate_kws = {}

    assert isinstance(annotate_kws, dict)

    _annotate_kws = {
        'ha':"center",
        "va": "center",
        "fmt": '%.2f',
        "textcolors": ("black", "white"),
        "threshold": None
    }

    _annotate_kws.update(annotate_kws)

    if annotate:
        annotate_imshow(im, values, **_annotate_kws)

    if yticklabels is not None:
        ax.set_yticks(np.arange(len(yticklabels)))
        ax.set_yticklabels(yticklabels)

    if xticklabels is not None:
        ax.set_xticks(np.arange(len(xticklabels)))
        if len(xticklabels) > 5:
            ax.set_xticklabels(xticklabels, rotation=70)
        ax.set_xticklabels(xticklabels)

    if ax_kws:
        process_axes(ax, **ax_kws)

    if grid_params:
        process_grid(ax, values, **grid_params)

    if colorbar:
        if cbar_params is None:
            cbar_params = {}
        process_cbar(ax, im, **cbar_params)
        # cb_tick_params = cb_tick_params or {'pad': 0.2, 'orientation': 'vertical'}
        # # https://stackoverflow.com/a/18195921/5982232
        # divider = make_axes_locatable(ax)
        # cax = divider.append_axes("right", size="5%", pad=0.2)
        # fig: plt.Figure = plt.gcf()
        # cb = fig.colorbar(im, cax=cax, **cb_tick_params)

    if show:
        plt.show()

    return im


def process_grid(
        ax:plt.Axes,
        data:np.ndarray,
        border:bool = False,
        color:str = "w",
        linewidth:Union[int, float] = 3,
        linestyle:str = '-'
):

    if not border:
        # Turn spines off and create white grid.
        if isinstance(ax.spines, dict):
            for sp in ax.spines:
                ax.spines[sp].set_visible(False)
        else:
            ax.spines[:].set_visible(False)

    ax.set_xticks(np.arange(data.shape[1] + 1) - .5, minor=True)
    ax.set_yticks(np.arange(data.shape[0] + 1) - .5, minor=True)
    ax.grid(which="minor", color=color, linestyle=linestyle, linewidth=linewidth)
    ax.tick_params(which="minor", bottom=False, left=False)
    return



__all__ = ["hist"]

from typing import Union, List

import numpy as np
import matplotlib.pyplot as plt

from .utils import kde
from .utils import process_axes, is_dataframe, create_subplots, is_series


def hist(
        x: Union[list, np.ndarray],
        labels:Union[str, List[str]] = None,
        share_axes:bool = True,
        grid: bool = True,
        subplots_kws:dict = None,
        add_kde:bool = False,
        kde_kws:dict = None,
        line_kws:dict = None,
        ax: plt.Axes = None,
        ax_kws: dict = None,
        show: bool = True,
        return_axes:bool = False,
        **kwargs
):
    """
    one stop shop for histogram

    Parameters
    -----------
        x : list, array, optional
            array like, numpy ndarray or pandas DataFrame, or list of arrays
        labels : list/str optional
            names of the arrays, used for setting the legend
        share_axes : bool (default=True)
            whether to draw all the histograms on one axes or not?
        grid : bool, optional
            whether to show the grid or not
        subplots_kws : dict
            kws which go to plt.subplots() such as figure size (width, height)
        add_kde: bool, (default=False)
            whether to add a line representing kernel densitiy estimation or not
        kde_kws : dict
            keyword arguments to calculate kde. These will go to :func:`easy_mpl.utils.kde`
            function.
        line_kws : dict
            keyword arguments for drawing the kde line. These will go to
            :obj:`matplotlib.axes.Axes.plot` function
        ax : plt.Axes, optional
            axes on which to draw the plot
        ax_kws : dict
            keyword arguments for :py:func:`easy_mpl.utils.process_axes`
        show : bool, optional
            whether to show the plot or not
        return_axes : bool, (default=False)
            whether to return the axes objects on which histogram/histograms are
            drawn or not. If True, then the function returns two objects, the
            first is a tuple (output of axes.hist) and second is the axes/list of axes
            on which histogram is drawn
        **kwargs : optional
            any keyword arguments for :obj:`matplotlib.axes.Axes.hist`

    Returns
    -------
        same what is returned by :obj:`matplotlib.axes.Axes.hist`

    Example
    --------
        >>> from easy_mpl import hist
        >>> import numpy as np
        >>> hist(np.random.random((10, 1)))

    See :ref:`sphx_glr_auto_examples_hist.py` for more examples

    .. _axes.hist:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.hist.html
    """

    if isinstance(x, np.ndarray):
        if len(x) == x.size:
            X = [x]
            names = [None]
        else:
            X = [x[:, i] for i in range(x.shape[1])]
            names = [f"{i}" for i in range(x.shape[1])]

    elif is_dataframe(x):
        X = []
        for col in x.columns:
            X.append(x[col].values)
        names = x.columns.tolist()

    elif is_series(x):
        X = x.values
        names = [x.name]

    elif isinstance(x, (list, tuple)) and isinstance(x[0], (list, tuple, np.ndarray)):
        assert all([len(np.array(array_like))==np.array(array_like).size for array_like in x]), f"""
        All arrays must be one dimensional."""
        X = [np.array(x_).reshape(-1,) for x_ in x]
        names = [None]*len(X)

    elif isinstance(x, (list, tuple)) and not is_dataframe(x[0]):
        X = [x]
        names = [None]
    else:
        raise ValueError(f"unrecognized type of x {type(x)}")

    if labels is not None:
        if isinstance(labels, str):
            labels = [labels]
        assert len(labels) == len(names), f"{len(names)} does not match data"
        names = labels

    if share_axes:
        nplots = 1
    else:
        nplots = len(names)

    if nplots == 1:
        share_axes = True

    if subplots_kws is None:
        subplots_kws = dict()

    if not kde_kws:
        kde_kws = dict()
    _line_kws = {'color': 'k'}
    if line_kws:
        _line_kws.update(line_kws)

    f, axes = create_subplots(nplots, ax=ax, **subplots_kws)

    if isinstance(axes, np.ndarray):
        axes = axes.flatten()

    outs = []

    for idx, x, name in zip(range(len(names)), X, names):
        if name:
            kwargs['label'] = name

        if add_kde:
            kwargs['density'] = True

        if share_axes:
            assert isinstance(axes, plt.Axes)
            ax = axes
        else:
            ax = axes[idx]

        out = ax.hist(x, **kwargs)
        outs.append(out)

        if add_kde:
            kde_kws['bins'] = len(out[0])
            add_kde_line(ax, x, kde_kws, _line_kws)

        _ax_kws = dict(grid=grid)
        if ax_kws:
            _ax_kws.update(ax_kws)
        process_axes(ax, **_ax_kws)

        if name:
            ax.legend()

    if show:
        plt.show()

    if len(outs)==1:
        outs = outs[0]

    if return_axes:
        return outs, axes

    return outs


def add_kde_line(axes:plt.Axes, data, kde_kws:dict, line_kws:dict):
    ind, y = kde(data, **kde_kws)
    axes.plot(ind, y, **line_kws)
    return




__all__ = ["plot"]

import numpy as np
import matplotlib.pyplot as plt

from .utils import process_axes, is_dataframe, is_series, create_subplots

# todo add share_axes argument

def plot(
        *args,
        share_axes:bool = True,
        show: bool = True,
        ax: plt.Axes = None,
        ax_kws:dict = None,
        **kwargs
) -> plt.Axes:
    """
    One liner plot function. It's use is not more complex than :obj:`matplotlib.axes.Axes.plot` or
    :obj:`matplotlib.pyplot.plot` . However it accomplishes all in one line what requires multiple
    lines in matplotlib. args and kwargs can be anything which goes into :obj:`matplotlib.pyplot.plot`
    or :obj:`matplotlib.axes.Axes.plot`.

    Parameters
    ----------
        *args :
            either a single array or x and y arrays or anything which can go to
            :obj:`matplotlib.axes.Axes.plot` or anything which can got to :obj:`matplotlib.pyplot.plot` .
        share_axes : bool (default=True)
            whether to draw all the plots on same axes or not. Only relevant for 2 dimensional data.
        ax : :obj:`matplotlib.axes`
            matplotlib axes object on which plot is to be drawn. If not given,
            then current active axes will be used.
        ax_kws : dict
            keyword arguments for :func:`easy_mpl.utils.process_axes`
        show : bool, optional (default=True)
            If set to True, plt.show() is called.
        **kwargs : optional
            Any keyword argument for :obj:`matplotlib.pyplot.plot` or :obj:`matplotlib.axes.Axes.plot`

    Returns
    -------
    :obj:`matplotlib.axes`
        :obj:`matplotlib.axes` on which the plot is drawn. If ``show`` is False, this axes
        can be used for further processing

    Example
    --------
        >>> from easy_mpl import plot
        >>> import numpy as np
        >>> plot(np.random.random(100))
        use x and y
        >>> plot(np.arange(100), np.random.random(100))
        use x and y
        >>> plot(np.arange(100), np.random.random(100), '.')
        string after arrays represent marker style
        >>> plot(np.random.random(100), '.')
        use cutom marker
        >>> plot(np.random.random(100), '--*')
        using label keyword
        >>> plot(np.random.random(100), '--*', label='label')
        log transform y-axis
        >>> plot(np.random.random(100), '--*', ax_kws={'logy':True}, label='label')

    See :ref:`sphx_glr_auto_examples_plot.py` for more examples

    """

    if ax_kws is None:
        ax_kws = {}

    plot_args = []

    marker = None
    if len(args) == 1:
        data, = args
        data = [data]
    elif len(args) == 2 and not isinstance(args[1], str):
        data = args
    elif len(args) == 2 and isinstance(args[1], str):
        data, marker = args[0], args[1]
        data = [data]
    elif len(args) == 3:
        *data, marker = args
        if isinstance(marker, np.ndarray):
            data.append(marker)
            marker = None
    else:
        data = args

    if marker:
        plot_args.append(marker)
        assert 'marker' not in kwargs  # we have alreay got marker

    s = data[0]


    nplots = 1
    if not share_axes and hasattr(s, 'shape') and s.shape[1] >1:
        nplots = s.shape[1]

    figsize = None
    if 'figsize' in ax_kws:
       figsize = ax_kws.pop('figsize')
    f, ax = create_subplots(nplots, ax=ax, figsize=figsize, ncols=1, sharex="all")
    if isinstance(ax, np.ndarray):
        ax = ax.flatten().tolist()
    elif isinstance(ax, plt.Axes):
        ax = [ax]

    if is_series(s):
        ax_kws['min_xticks'] = ax_kws.get('min_xticks', 3)
        ax_kws['max_xticks'] = ax_kws.get('max_xticks', 5)
        ax_kws['xlabel'] = ax_kws.get('xlabel', s.index.name)
        ax_kws['ylabel'] = ax_kws.get('ylabel', s.name)
    elif is_dataframe(s):
        ax_kws['min_xticks'] = ax_kws.get('min_xticks', 3)
        ax_kws['max_xticks'] = ax_kws.get('max_xticks', 5)
        if s.shape[1] == 1:
            ax_kws['xlabel'] = ax_kws.get('xlabel', s.index.name)
            ax_kws['ylabel'] = ax_kws.get('ylabel', s.columns.tolist()[0])
        else:
            if share_axes:
                ax = _plot_df_cols(ax[0], data, s, ax_kws, plot_args, kwargs)
            else:
                for idx, col in enumerate(s.columns):
                    #if 'label' not in kwargs:
                    kwargs['label'] = col
                    ax_kws['label'] = col
                    ax[idx].plot(s[col], *plot_args, **kwargs)

                    _process_axis(ax[idx], False, ax_kws)
            if show:
                plt.show()
            return ax

    elif hasattr(s, 'shape') and len(s.shape)>1 and s.shape[1]>1 and not share_axes:
        for idx, col in enumerate(range(s.shape[1])):
            ax[idx].plot(s[:, idx])
            _process_axis(ax[idx], False, ax_kws)
        if show:
            plt.show()
        return ax

    if 'label' in kwargs:
        ax_kws['label'] = kwargs['label']

    ax[0].plot(*data, *plot_args, **kwargs)
    return _process_axis(ax[0], show, ax_kws)


def _plot_df_cols(ax, data, s, ax_kws, plot_args, kwargs):
    ax_kws['xlabel'] = ax_kws.get('xlabel', s.index.name)
    ax_kws['label'] = ax_kws.get('label', s.columns.tolist())
    for col in s.columns:
        kwargs['label'] = col
        data[0] = s[col]
        ax.plot(*data, *plot_args, **kwargs)
    return process_axes(ax, **ax_kws)


def _process_axis(ax, show, kwargs):
    if kwargs:
        ax = process_axes(ax=ax, **kwargs)
    if kwargs.get('save', False):
        plt.savefig(f"{kwargs.get('name', 'fig.png')}")
    if show:
        plt.show()
    return ax





__all__ = ["ridge"]

import random
from typing import Union, List

import numpy as np
import matplotlib.pyplot as plt

from .utils import kde
from .utils import is_dataframe
from .utils import make_cols_from_cmap
from .utils import despine_axes


# colormaps for ridge plot
RIDGE_CMAPS = [
    "afmhot", "afmhot_r", "Blues", "bone",
    "BrBG", "BuGn", "coolwarm", "cubehelix",
    "gist_earth", "GnBu", "Greens", "magma",
    "ocean", "Pastel1", "pink", "PuBu", "PuBuGn",
    "RdBu", "Spectral",
]


def ridge(
        data: Union[np.ndarray, List[np.ndarray]],
        bw_method="scott",
        cut: float = 0.5,
        color: Union[str, List[str], np.ndarray, List[np.ndarray]] = None,
        fill_kws: dict = None,
        line_width: Union[int, List[int], float, List[float]] = 1.0,
        line_color: Union[str, List[str]] = "black",
        plot_kws: dict = None,
        xlabel: str = None,
        title: str = None,
        figsize: tuple = None,
        hspace: float = -0.7,
        labels: Union[str, List[str]] = None,
        share_axes: bool = False,
        ax: plt.Axes = None,
        show=True,
) -> List[plt.Axes,]:
    """
    plots distribution of features/columns/arrays in data as ridge.

    Parameters
    ----------
        data : array, DataFrame
            array or list of arrays or pandas DataFrame/Series
        bw_method : str, optional
        cut : float (default=0.5)
        color : str, optional
            color to fill the ridges. It can be any valid matplotlib color or color
            name or cmap name or a list of colors for each ridge.
        fill_kws : dict, (default=None)
            keyword arguments that will go to :obj:`matplotlib.axes.Axes.fill_between`
        line_width : int (default=1.0)
            with of line of ridges.
        line_color : str (default="black")
            color or colors of lines of ridges.
        plot_kws : dict optional
            any keyword argumenets that will go to :obj:`matplotlib.axes.Axes.plot` during plot of kde line
        xlabel : str, optional
            xlabel for the figure
        title : str, optional
            title of the figure
        figsize : tuple, optional
            size of figure
        hspace : float, optional (default=-0.7)
            amount of distance between plots
        labels : list/str (default=None
            names for xticks
        share_axes : bool, optional (default=False)
            whether to draw all ridges on same axes or separate axes
        ax : plt.Axes, optional (default=None)
            matplotlib axes object :obj:`matplotlib.axes` on which to draw the ridges.
             If given all ridges will be drawn on this axes.
        show : bool, optional
            whether to show the plot or not

    Returns
    -------
    list
        a list of :obj:`matplotlib.axes`

    Examples
    --------
    >>> import numpy as np
    >>> from easy_mpl import ridge
    >>> data_ = np.random.random((100, 3))
    >>> ridge(data_)
    ... # specifying colormap
    >>> ridge(data_, cmap="Blues")
    ... # using pandas DataFrame
    >>> import pandas as pd
    >>> ridge(pd.DataFrame(data_))

    See :ref:`sphx_glr_auto_examples_ridge.py` for more examples

    """

    if isinstance(data, np.ndarray):
        if len(data) == data.size:
            data = data.reshape(-1, 1)
        assert data.ndim == 2
        data = [data[:, i] for i in range(data.shape[1])]

    elif hasattr(data, 'name') and hasattr(data, 'values'):
        if labels is None:
            labels = [data.name]
        data = [data.values]
    elif is_dataframe(data):
        if labels is None:
            labels = data.columns.tolist()
        data = [data.values[:, i] for i in range(data.shape[1])]

    assert isinstance(data, list)

    if labels is None:
        labels =  [f"F{i}" for i in range(len(data))]

    n = len(data)

    if color is None:
        cmap = random.choice(RIDGE_CMAPS)
        # +2 because we want to ignore first and last in most cases
        colors = make_cols_from_cmap(cmap, n + 2)
    elif isinstance(color, str):
        if color in plt.colormaps():
            colors = make_cols_from_cmap(color, n + 2)
        else:
            colors = [color for _ in range(n)]
    elif isinstance(color, list):
        colors = color
    elif isinstance(color, np.ndarray):
        if len(color) == 3 and len(color) == color.size:
            colors = [color for _ in range(n)]
        else:
            colors = [color[:, i] for i in range(color.shape[1])]
    else:
        colors = color

    if plot_kws is None:
        plot_kws = dict()

    dist_maxes = {}
    xs = {}
    ys = {}
    for idx, col in enumerate(labels):
        ind, y = kde(data[idx], bw_method=bw_method, cut=cut)
        dist_maxes[col] = np.max(y)
        xs[col], ys[col] = ind, y

    ymaxes = dict(sorted(dist_maxes.items(), key=lambda item: item[1]))

    nrows = n
    if share_axes:
        nrows = 1

    if ax is None:
        fig, axes = plt.subplots(nrows, ncols=1,
                                 figsize=figsize or (10, 6),
                                 gridspec_kw={"hspace": hspace})
    else:
        share_axes = True
        axes = ax
        nrows = 1

    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])
    ax_objs = axes.tolist()

    if isinstance(line_color, str):
        line_color = [line_color for _ in range(len(ymaxes))]

    if isinstance(line_width, (int, float)):
        line_width = [line_width for _ in range(len(ymaxes))]

    for idx, col in enumerate(reversed(list(ymaxes.keys()))):

        ax_idx = idx
        if share_axes:
            ax_idx = 0

        if isinstance(colors[idx], str) and colors[idx] == "white":
            plot_kws['label'] = col

        ax = ax_objs[ax_idx]
        ax.plot(xs[col], ys[col],
                         color=line_color[idx],
                         linewidth=line_width[idx], **plot_kws)

        _fill_kws = {
            "alpha": 1.0,
            'where': np.array([True for _ in range(len(xs[col]))])
        }

        if fill_kws is None:
            fill_kws = dict()

        if isinstance(colors[idx], str) and colors[idx] == "white":
            pass
        else:
            fill_kws['label'] = col

        _fill_kws.update(fill_kws)

        ax.fill_between(xs[col], ys[col], color=colors[idx], **_fill_kws)

        # setting uniform y lims
        ax.set_ylim(0, max(ymaxes.values()))

        # make background transparent
        rect = ax.patch
        rect.set_alpha(0)

        if idx == n - 1:
            if xlabel:
                ax.set_xlabel(xlabel, fontsize=26, fontweight="bold")

            ax.tick_params(axis="x", labelsize=20)
        else:
            if nrows > 1:
                ax.set_xticklabels([])
                ax.set_xticks([])

        # remove borders, axis ticks, and labels
        if nrows > 1:
            ax.set_yticklabels([])
            ax.set_yticks([])
        else:
            ax.tick_params(axis="y", labelsize=20)

        despine_axes(ax)

        if not share_axes and nrows>1:
            ax.text(xs[col][0],
                             0.2,
                             col,
                             fontsize=20,
                             ha="right")
        idx += 1

    if not share_axes:
        plt.subplots_adjust(hspace=hspace)

    if title:
        plt.suptitle(title, fontsize=25)

    if show:
        if share_axes:
            plt.legend()

        plt.show()

    return ax_objs




__all__ = ["spider_plot"]

import math
from typing import Union, List, Optional

import numpy as np
import matplotlib.pyplot as plt

from .utils import register_projections


def spider_plot(
        data: Union[np.ndarray, list],
        tick_labels: list = None,
        highlight: Union[int, float] = None,
        plot_kws: Optional[Union[dict, List[dict]]] = None,
        xtick_kws: Optional[dict] = None,
        fill_kws: Optional[dict] = None,
        frame: str = "circle",
        color: Union[List[str], str] = None,
        fill_color: Union[List[str], str] = None,
        leg_kws: dict = None,
        labels: list = None,
        show: Optional[bool] = True,
        figsize=None,
)->plt.Axes:
    """
    Draws spider plot on an axes

    Parameters
    ----------
        data :
            values to display. It should be array like/numpy array or pandas DataFrame/Series
        tick_labels : list, optional (default=None)
            tick labels. It's length should be equal to length of values
        plot_kws : dict, optional (default=None)
            These can include, ``color``, ``linewidth``, ``linestyle`` etc
        xtick_kws : dict, optional (default=None)
            These can include ``color``, ``size`` etc.
        fill_kws : dict, optional
            These can include ``color``, ``alpha`` etc.
        highlight : int/float optional (default=None)
            whether to highlight a certain circular line or not
        color : str, optional (default=None)
            colormap to use
        frame: str, optional (default="circle")
            whether the outer frame and grids should be polygon or circle
        figsize : tuple, optional (default=None)
            figure size
        fill_color :
            color to use for filling
        leg_kws : dict
            keyword arguments that will go to ax.legend()
        labels: list, optional (default=None)
            the labels for values
        show : bool, optional (default=True)
            whether to show the plot or not

    Returns
    -------
    plt.Axes
       :obj:`matplotlib.axes` on which plot is drawn

    Examples
    --------
    >>> from easy_mpl import spider_plot
    >>> vals = [-0.2, 0.1, 0.0, 0.1, 0.2, 0.3]
    >>> spider_plot(values=vals)
    ... # specifying labels
    >>> labels = ['a', 'b','c', 'd', 'e', 'f']
    >>> spider_plot(values=vals, labels=labels)
    ... # specifying tick size
    >>> spider_plot(vals, labels, xtick_kws={'size': 13})
    ... # we can also pass dataframe
    >>> import pandas as pd
    >>> df = pd.DataFrame.from_dict(
    ... {'summer': {'a': -0.2, 'b': 0.1, 'c': 0.0, 'd': 0.1, 'e': 0.2, 'f': 0.3},
    ... 'winter': {'a': -0.3, 'b': 0.1, 'c': 0.0, 'd': 0.2, 'e': 0.15,'f': 0.25}})
    >>> spider_plot(df, xtick_kws={'size': 13})
    ... # use polygon frame
    >>> spider_plot(values=vals, frame="polygon")

    See :ref:`sphx_glr_auto_examples_spider_plot.py` for more examples

    """

    # todo, allow grids and frame to have different types

    if isinstance(data, (list, tuple)):
        data = np.array(data).reshape(-1, 1)

    if tick_labels is None:
        if hasattr(data, "index"):
            tick_labels = data.index
        else:
            tick_labels = [f"F{i}" for i in range(len(data))]

    if labels is None:
        if hasattr(data, "name"):
            labels = [data.name]
        elif hasattr(data, "columns"):
            labels = data.columns.tolist()
        else:
            labels = [f'Value_{i}' for i in range(data.shape[1])]

    if hasattr(data, "values"):
        data = data.values

    N = len(tick_labels)
    assert N == len(data)

    if not isinstance(plot_kws, list):
        plot_kws = [plot_kws for _ in range(data.shape[1])]

    def_color = plt.cm.get_cmap('Set2', data.shape[1])

    if color is None:
        color = [def_color(i) for i in range(data.shape[1])]
    elif not isinstance(color, list):
        color = [color for _ in range(data.shape[1])]

    if fill_color is None:
        fill_color = [def_color(i) for i in range(data.shape[1])]
    elif not isinstance(fill_color, list):
        fill_color = [fill_color for _ in range(data.shape[1])]

    angles = [n / float(N) * 2 * math.pi for n in range(N)]
    angles += angles[:1]

    if frame == "polygon":
        register_projections(N, frame)
        fig, ax = plt.subplots(figsize=figsize, nrows=1, ncols=1,
                                subplot_kw= dict(projection='radar'))
    else:
        ax = plt.subplot(polar=True)

    _xtick_kws = {'color': 'grey', 'size': 14}
    xtick_kws = xtick_kws or {}
    _xtick_kws.update(xtick_kws)

    plt.xticks(angles[:-1], tick_labels, **_xtick_kws)

    for idx in range(data.shape[1]):

        plot_kws_ = plot_kws[idx]

        _plot_kws = {"color": color[idx], "linewidth": 2, "linestyle": 'solid'}
        plot_kws_ = plot_kws_ or {}
        _plot_kws.update(plot_kws_)

        val = data[:, idx].tolist()
        val.append(val[0])
        ax.plot(angles, val, **_plot_kws)

        _fill_kws = {"color":fill_color[idx], "alpha":.4, 'label': '_nolegend_'}
        fill_kws = fill_kws or {}
        _fill_kws.update(fill_kws)
        ax.fill(angles, val, **_fill_kws)

    if highlight:
        val_minus = highlight - 0.005
        val_plus = highlight + 0.005
        ax.fill_between(np.linspace(0, 2 * np.pi, 100),
                        val_minus,
                        val_plus,
                        color='red',
                        zorder=10)

    plt.gca().set_rmax(np.max(data) + np.max(data)*0.2)
    plt.gca().set_rmin(np.min(data) - abs(np.min(data) * 0.2))

    _leg_kws = {'labelspacing': 0.1, 'fontsize': 12, 'bbox_to_anchor': (1.3, 1.1)}
    leg_kws = leg_kws or {}
    _leg_kws.update(leg_kws)
    legend = ax.legend(labels, **_leg_kws)

    if show:
        plt.tight_layout()
        plt.show()

    return ax




__all__ = ["violin_plot"]

import random
from typing import Union, List

import matplotlib
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rcParams

from easy_mpl.utils import _rescale, kde, is_dataframe


FILL_COLORS = [np.array([253,160,231])/255,
               np.array([102, 217, 191])/255,
               np.array([251, 173, 167])/255,
               np.array([220, 194, 102])/255,
               np.array([201, 185, 255])/255,
               np.array([205, 255, 205])/255,
               np.array([129, 207, 128])/255,
               np.array([103, 199, 209])/255,
               np.array([176, 204, 255])/255,
               ]


def violin_plot(
        data: Union[np.ndarray, List[np.ndarray]],
        X: Union[np.ndarray, List[np.ndarray]] = None,
        fill: bool = True,
        fill_colors=None,
        violin_kws: dict = None,
        show_datapoints: bool = True,
        datapoints_colors: Union[list, str] = None,
        scatter_kws: dict = None,
        show_boxplot: bool = False,
        box_kws: dict = None,
        label_violin: bool = False,
        index_method: str = "jitter",
        max_dots: Union[int, List[int]] = 100,
        cut: Union[float, tuple, List[float], List[tuple]] = 0.2,
        labels: Union[str, List[str]] = None,
        ax: plt.Axes = None,
        show: bool = True,
) -> plt.Axes:
    """
    makes violin plot/plots of the arrays in data

    parameters
    ----------
    data :
        It can array like or list of arrays. The length of each array need
        not be equal. If multiple arrays are given, then violin is
        drawn for each array
    X :
        indices for x-axes for each of the array in data. It can be
        array or list of arrays. If list of arrays is given, the length of
        arrays can be unequal.
    fill : bool, optional (default=True)
        whether to fill the violin with color or not
    fill_colors :
        colors to fill the violins
    violin_kws : dict (default=None)
        any keyword arugment for axes.violin_ in the form of dictionary
    show_datapoints : bool (default=True)
        whether to plot the datapoints or not
    datapoints_colors
        color for the datapoints/markers
    scatter_kws : dict (default=None)
        keyword arguments for axes.scatter_. This will only be valid if
        ``show_datapoints`` is True.
    show_boxplot : bool (default=False)
        whether to show the boxplot inside the voilin or not?
    box_kws : dict (default=None)
        keyword arguments for :obj:`matplotlib.axes.Axes.boxplot`. This will only be valid if
        ``show_boxplot`` is True.
    label_violin : bool (default=False)
        whether to label mean value of each violin or not
    index_method : str (default="jitter")
        Only valid if `X` is not given. The method to generate indices for x-axis.
        See `this <https://stackoverflow.com/a/33965400/5982232>_` for context
    max_dots : int/list (default=100)
        maximum number of dots to show. It can also be a list of integers, which
        would define the number of dots for each array in X.
    cut : float/list (default=0.2)
        This variables determines the length of violin. If given as a list or
        list of tuples, it should match the number of arrays in X.
    labels : list/str (default=None
        names for xticks
    ax : plt.Axes (default=None)
        matplotlib Axes object :obj:`matplotlib.axes` on which to draw the plot. If not given, then
        the currently available axes from plt.gca will be used.
    show : bool (default=True)
        whether to show the plot or not

    Returns
    -------
    plt.Axes
        the matplotlib Axes object on which violin/violins are drawn

    Examples
    --------
    >>> import numpy as np
    >>> from easy_mpl import violin_plot
    >>> data = np.random.gamma(20, 10, 100)
    >>> violin_plot(data)
    >>> violin_plot(data, show_datapoints=False)
    >>> violin_plot(data, show_datapoints=False, show_boxplot=True)

    See :ref:`sphx_glr_auto_examples_violin.py` for more examples

    .. _axes.scatter:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.scatter.html

    .. _axes.violin:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.violin.html
    """

    if isinstance(data, np.ndarray):
        Y = to_1d_arrays(data)

    elif is_dataframe(data):   # data is pd.DataFrame
        Y = to_1d_arrays(data.values)
        if labels is None:
            labels = data.columns.tolist()
    else:
        assert isinstance(data, list), f"Unrecognized data of type {data.__class__.__name__}"
        Y = data

    if X is None:
        if index_method == "jitter":
            X = jittered_ind(Y)
            offsets = [0] * len(Y)
        else:
            maxs = get_maxs(Y, cut=cut)
            X, offsets = [], []
            for idx, y in enumerate(Y):
               x = beeswarm_ind(y, nbins=100, lim=maxs[idx]*0.8)
               X.append(x)
               offsets.append(idx)
    else:
        offsets = [0] * len(Y)


    if not isinstance(X, list):
        X = [X]

    assert isinstance(X, list)
    assert len(X) == len(Y)

    # Horizontal positions for the violins.
    # They are arbitrary numbers. They could have been [-1, 0, 1] for example.
    POSITIONS = range(len(X))

    if ax is None:
        ax = plt.gca()

    # Add violins ----------------------------------------------------
    # bw_method="silverman" means the bandwidth of the kernel density
    # estimator is computed via Silverman's rule of thumb.
    # More on this in the bonus track ;)

    _violin_kws = {
        "widths":0.45,
        "bw_method": "silverman",
        "showmeans": False,
        "showmedians": False,
        "showextrema": False
    }

    if violin_kws is None:
        violin_kws = dict()

    _violin_kws.update(violin_kws)

    vpstats = get_vpstats(Y,
                          bw_method=_violin_kws.get('bw_method', None),
                          cut=cut)

    if "bw_method" in _violin_kws:
        _violin_kws.pop("bw_method")

    # The output is stored in 'violins', used to customize their appearence
    if matplotlib.__version__ <= "3.4.0":
        violins = violin_mpl_340(ax, vpstats, positions=POSITIONS, **_violin_kws)
    else:
        violins = ax.violin(vpstats, positions=POSITIONS, **_violin_kws)

    if isinstance(fill_colors, str):
        fill_colors = [fill_colors for _ in range(len(X))]

    if fill_colors is None:
        fill_colors = random.choices(FILL_COLORS, k=len(X))

    # Customize violins (remove fill, customize line, etc.)
    for idx, pc in enumerate(violins["bodies"]):
        if fill:
            pc.set_facecolor(fill_colors[idx])
        else:
            pc.set_facecolor("none")
        pc.set_edgecolor("#282724")
        pc.set_alpha(1)

    # Add boxplots
    if show_boxplot:

        boxprops = dict(
            linewidth=2,
            color="#747473"  # GREY_DARK
        )

        _box_kws = {
            'medianprops': dict(
                linewidth=4,
                color="#747473",  # GREY_DARK
                solid_capstyle="butt"
            ),
            'boxprops': boxprops,
            'whiskerprops': boxprops,
            'showfliers': False,  # Do not show the outliers beyond the caps.
            'showcaps': False,  # Do not show the caps
        }

        if box_kws is None:
            box_kws = dict()

        _box_kws.update(box_kws)
        ax.boxplot(Y, positions=POSITIONS, **_box_kws)

    # Add  dots
    if show_datapoints:
        _scatter_kws = {
            'alpha': 0.4,
            's': 10
        }

        if scatter_kws is None:
            scatter_kws = dict()

        _scatter_kws.update(scatter_kws)

        if isinstance(datapoints_colors, str):
            datapoints_colors = [datapoints_colors for _ in range(len(X))]

        if datapoints_colors is None:
            datapoints_colors = ["black" for _ in range(len(X))]

        if isinstance(max_dots, int):
            max_dots = [max_dots for _ in range(len(X))]

        for offset, x, y, color, n_dots in zip(offsets, X, Y, datapoints_colors, max_dots):

            x, y = sample(x+offset, y, n=n_dots)

            ax.scatter(x, y, color=color, **_scatter_kws)

    if label_violin:
        # Add mean value labels ------------------------------------------
        means = [y.mean() for y in Y]
        _mean_kws = {
            's': 12,
            'color': '#850e00', # dark red
            'zorder': 3
        }
        for i, mean in enumerate(means):
            # Add dot representing the mean
            ax.scatter(i, mean, **_mean_kws)

            # Add line conecting mean value and its label
            ax.plot([i, i + 0.25], [mean, mean], ls="dashdot", color="black", zorder=3)

            # Add mean value label.
            ax.text(
                i + 0.25,
                mean,
                r"$\hat{\mu}_{\rm{mean}} = $" + str(round(mean, 2)),
                fontsize=13,
                va="center",
                bbox=dict(
                    facecolor="white",
                    edgecolor="black",
                    boxstyle="round",
                    pad=0.15
                ),
                zorder=10  # to make sure the line is on top
            )

    if labels is not None:
        kws = dict()
        if len(labels)>7:
            kws['rotation'] = 90
        ax.set_xticks(range(len(labels)))
        ax.set_xticklabels(labels, **kws)

    if show:
        plt.show()

    return ax


def jittered_ind(y, jitter=0.04):
    import scipy.stats as st

    x_data = [np.array([i] * len(d)) for i, d in enumerate(y)]
    return [x + st.t(df=6, scale=jitter).rvs(len(x)) for x in x_data]


def beeswarm_ind(y, nbins=None, lim=1.0):
    """
    Returns x coordinates for the points in ``y``, so that plotting ``x`` and
    ``y`` results in a bee swarm plot.
    """
    y = np.asarray(y)
    if nbins is None:
        nbins = len(y) // 6

    # Get upper bounds of bins
    ind = np.zeros(len(y))
    ylo = np.min(y)
    yhi = np.max(y)
    dy = (yhi - ylo) / nbins
    ybins = np.linspace(ylo + dy, yhi - dy, nbins - 1)

    # Divide indices into bins
    i = np.arange(len(y))
    ibs = [0] * nbins
    ybs = [0] * nbins
    nmax = 0
    for j, ybin in enumerate(ybins):
        f = y <= ybin
        ibs[j], ybs[j] = i[f], y[f]
        nmax = max(nmax, len(ibs[j]))
        f = ~f
        i, y = i[f], y[f]
    ibs[-1], ybs[-1] = i, y
    nmax = max(nmax, len(ibs[-1]))

    # Assign x indices
    dx = 1 / (nmax // 2)
    for i, y in zip(ibs, ybs):
        if len(i) > 1:
            j = len(i) % 2
            i = i[np.argsort(y)]
            a = i[j::2]
            b = i[j+1::2]
            ind[a] = (0.5 + j / 3 + np.arange(len(b))) * dx
            ind[b] = (0.5 + j / 3 + np.arange(len(b))) * -dx

    return _rescale(ind, -lim, lim)


def get_vpstats(dataset, bins=100, cut=0.2, bw_method=None):

    if isinstance(cut, (float, tuple)):
        cut = [cut for _ in range(len(dataset))]

    assert isinstance(cut, list)
    assert len(cut)==len(dataset)

    vpstats = []

    for idx, y in enumerate(dataset):
        stats = dict()
        coords, vals = kde(y, bw_method=bw_method, bins=bins, cut=cut[idx])
        stats['coords'] = coords
        stats['vals'] = vals
        stats['mean'] = np.mean(y)
        stats['median'] = np.median(y)
        stats['min'] = np.min(y)
        stats['max'] = np.max(y)
        vpstats.append(stats)

    return vpstats


def get_maxs(dataset, points=100, quantiles=None, bw_method=None, cut=0.5):

    vpstats = get_vpstats(dataset,
                          bins=points, #quantiles=quantiles,
                          bw_method=bw_method,
                          cut=cut)

    N = len(vpstats)
    widths = [0.45] * N

    mins, maxs = [], []
    for stats, width in zip(vpstats, widths):

        vals = np.array(stats['vals'])
        vals = 0.5 * width * vals / vals.max()
        #mins.append(np.min(vals))
        maxs.append(np.max(vals))

    return maxs


def sample(x, y, n=100):
    if len(x)<=n:
        return x, y
    idx = np.random.choice(np.arange(len(x)), n, replace=False)

    return x[idx], y[idx]


def to_1d_arrays(data)->List[np.ndarray,]:
    if len(data) == data.size:
        Y = [data.reshape(-1,)]
    else:
        assert data.ndim == 2
        Y = [data[:, i] for i in range(data.shape[1])]
    return Y

# ******************************************
# Following code is taken from matplotlib
# because for versions prior to 3.5 (3.4 or prior versions)
# we need to set where argument to fill_between otherwise
# fill_between returns error.
# matplotlib comes with following licence
"""
License agreement for matplotlib versions 1.3.0 and later
=========================================================

1. This LICENSE AGREEMENT is between the Matplotlib Development Team
("MDT"), and the Individual or Organization ("Licensee") accessing and
otherwise using matplotlib software in source or binary form and its
associated documentation.

2. Subject to the terms and conditions of this License Agreement, MDT
hereby grants Licensee a nonexclusive, royalty-free, world-wide license
to reproduce, analyze, test, perform and/or display publicly, prepare
derivative works, distribute, and otherwise use matplotlib
alone or in any derivative version, provided, however, that MDT's
License Agreement and MDT's notice of copyright, i.e., "Copyright (c)
2012- Matplotlib Development Team; All Rights Reserved" are retained in
matplotlib  alone or in any derivative version prepared by
Licensee.

3. In the event Licensee prepares a derivative work that is based on or
incorporates matplotlib or any part thereof, and wants to
make the derivative work available to others as provided herein, then
Licensee hereby agrees to include in any such work a brief summary of
the changes made to matplotlib .

4. MDT is making matplotlib available to Licensee on an "AS
IS" basis.  MDT MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, MDT MAKES NO AND
DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF MATPLOTLIB
WILL NOT INFRINGE ANY THIRD PARTY RIGHTS.

5. MDT SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF MATPLOTLIB
 FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR
LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING
MATPLOTLIB , OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF
THE POSSIBILITY THEREOF.

6. This License Agreement will automatically terminate upon a material
breach of its terms and conditions.

7. Nothing in this License Agreement shall be deemed to create any
relationship of agency, partnership, or joint venture between MDT and
Licensee.  This License Agreement does not grant permission to use MDT
trademarks or trade name in a trademark sense to endorse or promote
products or services of Licensee, or any third party.

8. By copying, installing or otherwise using matplotlib ,
Licensee agrees to be bound by the terms and conditions of this License
Agreement.
"""


def violin_mpl_340(self, vpstats, positions=None, vert=True, widths=0.5,
           showmeans=False, showextrema=True, showmedians=False):
    """
    Drawing function for violin plots.

    Draw a violin plot for each column of *vpstats*. Each filled area
    extends to represent the entire data range, with optional lines at the
    mean, the median, the minimum, the maximum, and the quantiles values.

    Parameters
    ----------
    vpstats : list of dicts
      A list of dictionaries containing stats for each violin plot.
      Required keys are:

      - ``coords``: A list of scalars containing the coordinates that
        the violin's kernel density estimate were evaluated at.

      - ``vals``: A list of scalars containing the values of the
        kernel density estimate at each of the coordinates given
        in *coords*.

      - ``mean``: The mean value for this violin's dataset.

      - ``median``: The median value for this violin's dataset.

      - ``min``: The minimum value for this violin's dataset.

      - ``max``: The maximum value for this violin's dataset.

      Optional keys are:

      - ``quantiles``: A list of scalars containing the quantile values
        for this violin's dataset.

    positions : array-like, default: [1, 2, ..., n]
      The positions of the violins. The ticks and limits are
      automatically set to match the positions.

    vert : bool, default: True.
      If true, plots the violins vertically.
      Otherwise, plots the violins horizontally.

    widths : array-like, default: 0.5
      Either a scalar or a vector that sets the maximal width of
      each violin. The default is 0.5, which uses about half of the
      available horizontal space.

    showmeans : bool, default: False
      If true, will toggle rendering of the means.

    showextrema : bool, default: True
      If true, will toggle rendering of the extrema.

    showmedians : bool, default: False
      If true, will toggle rendering of the medians.

    Returns
    -------
    dict
      A dictionary mapping each component of the violinplot to a
      list of the corresponding collection instances created. The
      dictionary has the following keys:

      - ``bodies``: A list of the `~.collections.PolyCollection`
        instances containing the filled area of each violin.

      - ``cmeans``: A `~.collections.LineCollection` instance that marks
        the mean values of each of the violin's distribution.

      - ``cmins``: A `~.collections.LineCollection` instance that marks
        the bottom of each violin's distribution.

      - ``cmaxes``: A `~.collections.LineCollection` instance that marks
        the top of each violin's distribution.

      - ``cbars``: A `~.collections.LineCollection` instance that marks
        the centers of each violin's distribution.

      - ``cmedians``: A `~.collections.LineCollection` instance that
        marks the median values of each of the violin's distribution.

      - ``cquantiles``: A `~.collections.LineCollection` instance created
        to identify the quantiles values of each of the violin's
        distribution.

    """

    # Statistical quantities to be plotted on the violins
    means = []
    mins = []
    maxes = []
    medians = []
    quantiles = np.asarray([])

    # Collections to be returned
    artists = {}

    N = len(vpstats)
    datashape_message = ("List of violinplot statistics and `{0}` "
                         "values must have the same length")

    # Validate positions
    if positions is None:
        positions = range(1, N + 1)
    elif len(positions) != N:
        raise ValueError(datashape_message.format("positions"))

    # Validate widths
    if np.isscalar(widths):
        widths = [widths] * N
    elif len(widths) != N:
        raise ValueError(datashape_message.format("widths"))

    # Calculate ranges for statistics lines
    pmins = -0.25 * np.array(widths) + positions
    pmaxes = 0.25 * np.array(widths) + positions

    # Check whether we are rendering vertically or horizontally
    if vert:
        fill = self.fill_betweenx
        perp_lines = self.hlines
        par_lines = self.vlines
    else:
        fill = self.fill_between
        perp_lines = self.vlines
        par_lines = self.hlines

    if rcParams['_internal.classic_mode']:
        fillcolor = 'y'
        edgecolor = 'r'
    else:
        fillcolor = edgecolor = self._get_lines.get_next_color()

    # Render violins
    bodies = []
    for stats, pos, width in zip(vpstats, positions, widths):
        # The 0.5 factor reflects the fact that we plot from v-p to
        # v+p
        vals = np.array(stats['vals'])
        vals = 0.5 * width * vals / vals.max()
        bodies += [fill(stats['coords'],
                        -vals + pos,
                        vals + pos,
                        facecolor=fillcolor,
                        where = np.array([True for _ in range(len(stats['coords']))]),
                        alpha=0.3)]
        means.append(stats['mean'])
        mins.append(stats['min'])
        maxes.append(stats['max'])
        medians.append(stats['median'])
        q = stats.get('quantiles')
        if q is not None:
            # If exist key quantiles, assume it's a list of floats
            quantiles = np.concatenate((quantiles, q))
    artists['bodies'] = bodies

    # Render means
    if showmeans:
        artists['cmeans'] = perp_lines(means, pmins, pmaxes,
                                       colors=edgecolor)

    # Render extrema
    if showextrema:
        artists['cmaxes'] = perp_lines(maxes, pmins, pmaxes,
                                       colors=edgecolor)
        artists['cmins'] = perp_lines(mins, pmins, pmaxes,
                                      colors=edgecolor)
        artists['cbars'] = par_lines(positions, mins, maxes,
                                     colors=edgecolor)

    # Render medians
    if showmedians:
        artists['cmedians'] = perp_lines(medians,
                                         pmins,
                                         pmaxes,
                                         colors=edgecolor)

    # Render quantile values
    if quantiles.size > 0:
        # Recalculate ranges for statistics lines for quantiles.
        # ppmins are the left end of quantiles lines
        ppmins = np.asarray([])
        # pmaxes are the right end of quantiles lines
        ppmaxs = np.asarray([])
        for stats, cmin, cmax in zip(vpstats, pmins, pmaxes):
            q = stats.get('quantiles')
            if q is not None:
                ppmins = np.concatenate((ppmins, [cmin] * np.size(q)))
                ppmaxs = np.concatenate((ppmaxs, [cmax] * np.size(q)))
        # Start rendering
        artists['cquantiles'] = perp_lines(quantiles, ppmins, ppmaxs,
                                           colors=edgecolor)

    return artists



from setuptools import setup

import os

long_desc="easy_mpl"

fpath = os.path.join(os.getcwd(), "readme.md")
if os.path.exists(fpath):
    with open(fpath, "r") as fd:
        long_desc = fd.read()

setup(

    name='easy_mpl',

    version="0.21.3",

    description='one stop shop for matplotlib plots',
    long_description=long_desc,
    long_description_content_type="text/markdown",

    url='https://github.com/Sara-Iftikhar/easy_mpl',

    author='Sara Iftikhar',
    author_email='sara.rwpk@gmail.com',

    classifiers=[
        'Development Status :: 4 - Beta',

        'Natural Language :: English',

        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    packages=['easy_mpl'],

    # 3.6.0 version has problem with parallel coords
    install_requires=['matplotlib>=3.0.0, <4.0.0',
                       ],
    extras_require = {"all": [
        "numpy>=1.16.2, <2.0.0",
        "matplotlib>=3.0.0, <4.0.0",
        "pandas>=0.23.3, <2.0.0",
        "scipy"]}
)